from dataclasses import replace
from enum import auto
from pathlib import Path
from typing import List

import typer
from pynmrstar import Entry
from strenum import LowercaseStrEnum

from nef_pipelines.lib.nef_lib import (
    UNUSED,
    read_entry_from_file_or_stdin_or_exit_error,
    select_frames,
)
from nef_pipelines.lib.sequence_lib import sequence_from_entry_or_exit
from nef_pipelines.lib.structures import AtomLabel, Residue
from nef_pipelines.lib.util import STDIN, is_int, parse_comma_separated_options
from nef_pipelines.tools.frames import frames_app


class AssignmentState(LowercaseStrEnum):
    PARTIAL = auto()
    FULL = auto()


# TODO should be able decide if partial assignments are required and if the residue type need to be set
# also we do nothing with pseudo residues...
# also we don't check the atom name is compatible with the residue or chem_comp
# obvious we could filter on other fields such as values

FRAMES_HELP = """\
    the names of the  frames to use, this can be a comma separated list of name or the option can be called
    called multiple times. Wild cards are allowed. If no match is found wild cards are checked as well unless
    --exact is set
"""

ASSIGNED_HELP = """
    if set lines which are consider assigned [chain_code, sequence_code and atom_name set]
    are deleted, otherwise unassigned lines are deleted. Note the --state option controls
    if all assignments on need to be complete for the line to be considered assigned
"""

STATE_HELP = """
    how complete do the assignments in a row have to be for it to be considered assigned, the default depends on
    the option --assigned if assigned is false the value is full, else it is partial
"""


# rename from filter when api is changed, filter shadows a builtin
@frames_app.command()
def filter(
    frame_selectors: List[str] = typer.Option(["*"], help=FRAMES_HELP),
    input_path: Path = typer.Option(
        STDIN,
        "--in",
        "-i",
        metavar="|INPUT|",
        help="input to read NEF data from [stdin = -]",
    ),
    filter_assigned: bool = typer.Option(
        False,
        "--assigned",
        help=ASSIGNED_HELP,
    ),
    assignment_state: AssignmentState = typer.Option(None, "--state", help=STATE_HELP),
):
    """-  filter the lines in one or more save frames depending on assignment status
    [currently chain_code, sequence_code and atom_name must be set and the sequence code must be in molecular
    system for an atom to be assigned]
    """

    frame_selectors = parse_comma_separated_options(frame_selectors)

    if assignment_state is None:
        if filter_assigned:
            assignment_state = AssignmentState.PARTIAL
        else:
            assignment_state = AssignmentState.FULL

    entry = read_entry_from_file_or_stdin_or_exit_error(input_path)

    entry = pipe(entry, frame_selectors, filter_assigned, assignment_state)

    print(entry)


def pipe(
    entry: Entry,
    frame_selectors: List[str],
    filter_assigned=False,
    target_assignment_state=AssignmentState.FULL,
) -> Entry:

    sequence = sequence_from_entry_or_exit(entry)

    sequence_set = set([Residue.from_sequence_residue(residue) for residue in sequence])
    # currently we ignore residue types
    sequence_set = {replace(residue, residue_name="") for residue in sequence_set}

    selected_frames = select_frames(entry, frame_selectors)
    for frame in selected_frames:

        for loop in frame.loops:

            rows_to_remove = set()
            tags = loop.tags

            loop_has_tags = {}
            for tag_name in (
                "chain_code",
                "sequence_code",
                "residue_name",
                "atom_name",
            ):
                loop_has_tags[tag_name] = [tag.startswith(tag_name) for tag in tags]

            are_assignables = [any(has_tag) for has_tag in loop_has_tags.values()]
            is_assignable = all(are_assignables)

            if is_assignable:

                tag_sets_by_end = {}

                for tag_name in (
                    "chain_code",
                    "sequence_code",
                    "residue_name",
                    "atom_name",
                ):
                    for tag, is_chain_code in zip(tags, loop_has_tags[tag_name]):
                        if is_chain_code:
                            end = tag[len(tag_name) :]
                            tag_and_index = (tag, loop.tag_index(tag))
                            tag_sets_by_end.setdefault(end, []).append(tag_and_index)

                for row_index, row in enumerate(loop):

                    assignment_state = _get_assignment_state(
                        row, sequence_set, tag_sets_by_end
                    )

                    assigned = False
                    if target_assignment_state is AssignmentState.FULL:
                        assigned = assignment_state is AssignmentState.FULL
                    elif target_assignment_state is AssignmentState.PARTIAL:
                        assigned = assignment_state in {
                            AssignmentState.FULL,
                            AssignmentState.PARTIAL,
                        }

                    if filter_assigned and assigned:
                        rows_to_remove.add(row_index)
                    elif not filter_assigned and not assigned:
                        rows_to_remove.add(row_index)

            for row_index in reversed(sorted(rows_to_remove)):
                del loop.data[row_index]
    return entry


def _get_assignment_state(row, sequence_set, tag_sets_by_end):
    atom_labels_by_tag_set_end = _get_atom_labels_by_tag_set(row, tag_sets_by_end)

    assigned_by_end = {}
    for set_end, atom_label in atom_labels_by_tag_set_end.items():
        residue_in_sequence = atom_label.residue in sequence_set
        atom_name = atom_label.atom_name
        atom_name_ok = (
            atom_name is not None and len(atom_name) > 0 and atom_name != UNUSED
        )

        assigned_by_end[set_end] = residue_in_sequence and atom_name_ok

    result = None
    if all(assigned_by_end.values()):
        result = AssignmentState.FULL
    elif any(assigned_by_end.values()):
        result = AssignmentState.PARTIAL

    return result


def _get_atom_labels_by_tag_set(row, tag_sets_by_end):

    residues_by_tag_end = {}
    for tag_end, tag_set in tag_sets_by_end.items():

        chain_code = None
        sequence_code = None
        # residue_name = None
        atom_name = None

        for tag, tag_index in tag_set:

            value = row[tag_index]

            if tag.startswith("chain_code"):
                chain_code = value
            elif tag.startswith("sequence_code"):
                sequence_code = value
            # ignore this for now
            # elif tag.startswith("residue_name"):
            #     residue_name = value
            elif tag.startswith("atom_name"):
                atom_name = value

            if is_int(sequence_code):
                sequence_code = int(sequence_code)

            residue = Residue(
                chain_code=chain_code,
                sequence_code=sequence_code,
                residue_name="",
            )
            atom_label = AtomLabel(residue, atom_name)
            residues_by_tag_end[tag_end] = atom_label

    return residues_by_tag_end



import datetime
from random import randint

from pynmrstar import Loop, Saveframe

from nef_pipelines.lib.constants import NEF_PIPELINES, NEF_VERSION


def get_creation_time():
    return datetime.datetime.now().isoformat()


def get_uuid(name, creation_time):
    random_value = "".join(["{}".format(randint(0, 9)) for _ in range(10)])
    return f"{name}-{creation_time}-{random_value}"


def create_header_frame(program_name, program_version, script_name):
    frame = Saveframe.from_scratch("nef_nmr_meta_data", "nef_nmr_meta_data")

    frame.add_tag("sf_category", "nef_nmr_meta_data")
    frame.add_tag("sf_framecode", "nef_nmr_meta_data")
    frame.add_tag("format_name", "nmr_exchange_format")
    frame.add_tag("nef_nmr_meta_data.format_version", NEF_VERSION)
    frame.add_tag("program_name", program_name)
    frame.add_tag("script_name", script_name)
    frame.add_tag("program_version", program_version)

    creation_date = get_creation_time()
    uuid = get_uuid(NEF_PIPELINES, creation_date)
    frame.add_tag("creation_date", creation_date)
    frame.add_tag("uuid", uuid)

    loop = Loop.from_scratch("nef_run_history")
    frame.add_loop(loop)

    history_tags = "run_number", "program_name", "program_version", "script_name"
    loop.add_tag(history_tags)

    return frame



import typer

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    isolate_loop,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.xeasy.importers.sequence import sequence

app = typer.Typer()
app.command()(sequence)

EXPECTED = """\
     loop_
      _nef_sequence.index
      _nef_sequence.chain_code
      _nef_sequence.sequence_code
      _nef_sequence.residue_name
      _nef_sequence.linking
      _nef_sequence.residue_variant
      _nef_sequence.cis_peptide

     1   A   1   HIS   start    .   .
     2   A   2   MET   middle   .   .
     3   A   3   ARG   middle   .   .
     4   A   4   GLN   middle   .   .
     5   A   5   THR   middle   .   .
     6   A   6   MET   middle   .   .
     7   A   7   LEU   middle   .   .
     8   A   8   VAL   middle   .   .
     9   A   9   THR   end      .   .

   stop_

"""


def test_basic():

    sequence_path = path_in_test_data(__file__, "basic.seq")

    result = run_and_report(app, [sequence_path])

    assert_lines_match(
        EXPECTED,
        isolate_loop(result.stdout, "nef_molecular_system", "nef_sequence"),
    )



from textwrap import dedent

import typer
from pynmrstar import Entry

from nef_pipelines import nef_app
from nef_pipelines.lib.globals_lib import get_global, set_global
from nef_pipelines.lib.nef_lib import (
    UNUSED,
    _parse_globals,
    create_entry_from_stdin,
    create_nef_save_frame,
)

VERBOSE_HELP = """\
    display more verbose information about NEF-Pipelines as it runs
"""

FORCE_HELP = """\
    globally force overwrite of output files even if it they exist and aren't empty
"""

INFO = """
    this is a frame for storing global options inside NEF-Pipelines, it can be safely deleted,
    [you can delete it automatically if you use 'nef save -' as the last command in your pipeline]
"""

INFO = dedent(INFO)


INFO = INFO.strip()
INFO = dedent(INFO)
INFO = INFO.replace(r"\s+", " ")

if nef_app:
    # noinspection PyUnusedLocal
    @nef_app.app.command("globals")
    def globals_(
        verbose: bool = typer.Option(None, "-v", "--verbose", help=VERBOSE_HELP),
        force: bool = typer.Option(None, "-f", "--force", help=FORCE_HELP),
    ):
        "- add global options to the pipeline [use save as your last command to clean up the globals]"

        entry = create_entry_from_stdin()
        if entry:
            _parse_globals(entry)

        if not entry and verbose or force:
            entry = Entry.from_scratch("nefpls_globals")

        if verbose is None:
            verbose = False
        if force is None:
            force = False

        set_global("force", force)
        set_global("verbose", verbose)
        set_global("info", INFO)

        entry = _create_or_update_globals_frame(entry)

        if entry:
            print(entry)
        else:
            print()


def _create_or_update_globals_frame(entry):

    if entry:
        global_frames = entry.get_saveframes_by_category("nefpls_globals")
        for global_frame in global_frames:
            entry.remove_saveframe(global_frame)

        global_frame = create_nef_save_frame("nefpls_globals")
        entry.add_saveframe(global_frame)

        tags = "force verbose info".split()

        for name in tags:
            global_frame.add_tags([(name, get_global(name, UNUSED))])

    return entry



from textwrap import dedent

import pytest
import typer
from pynmrstar import Entry
from typer.testing import CliRunner

from nef_pipelines.lib.nef_lib import NEF_MOLECULAR_SYSTEM
from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    isolate_frame,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.lib.util import is_int
from nef_pipelines.transcoders.fasta.importers.sequence import sequence

runner = CliRunner()
app = typer.Typer()
app.command()(sequence)


EXPECTED_3AA = """\
save_nef_molecular_system
   _nef_molecular_system.sf_category   nef_molecular_system
   _nef_molecular_system.sf_framecode  nef_molecular_system

   loop_
      _nef_sequence.index
      _nef_sequence.chain_code
      _nef_sequence.sequence_code
      _nef_sequence.residue_name
      _nef_sequence.linking
      _nef_sequence.residue_variant
      _nef_sequence.cis_peptide

     1   A   1   ALA   start    .   .
     2   A   2   ALA   middle   .   .
     3   A   3   ALA   end      .   .

   stop_

save_"""


# noinspection PyUnusedLocal
def test_3aa():

    path = path_in_test_data(__file__, "3aa.fasta")
    result = run_and_report(app, [path])

    assert Entry.from_string(result.stdout).entry_id == "test"
    mol_sys_result = isolate_frame(result.stdout, "%s" % NEF_MOLECULAR_SYSTEM)

    assert_lines_match(EXPECTED_3AA, mol_sys_result)


def test_3aa_spaces():

    path = path_in_test_data(__file__, "3aa_spaces.fasta")
    result = run_and_report(app, [path])

    assert Entry.from_string(result.stdout).entry_id == "fasta"
    mol_sys_result = isolate_frame(result.stdout, "%s" % NEF_MOLECULAR_SYSTEM)

    assert_lines_match(EXPECTED_3AA, mol_sys_result)


EXPECTED_3A_AB = """\
save_nef_molecular_system
   _nef_molecular_system.sf_category   nef_molecular_system
   _nef_molecular_system.sf_framecode  nef_molecular_system

   loop_
      _nef_sequence.index
      _nef_sequence.chain_code
      _nef_sequence.sequence_code
      _nef_sequence.residue_name
      _nef_sequence.linking
      _nef_sequence.residue_variant
      _nef_sequence.cis_peptide

     1   A    1   ALA   start    .   .
     2   A    2   ALA   middle   .   .
     3   A    3   ALA   end      .   .
     4   B    1   ALA   start    .   .
     5   B    2   ALA   middle   .   .
     6   B    3   ALA   end      .   .

   stop_

save_"""


# noinspection PyUnusedLocal
def test_3aa_x2():

    path = path_in_test_data(__file__, "3aa_x2.fasta")
    result = run_and_report(app, [path])

    assert Entry.from_string(result.stdout).entry_id == "test__test2"
    mol_sys_result = isolate_frame(result.stdout, "%s" % NEF_MOLECULAR_SYSTEM)

    assert_lines_match(EXPECTED_3A_AB, mol_sys_result)


EXPECTED_3A_AB_B_start_11 = """\
save_nef_molecular_system
   _nef_molecular_system.sf_category   nef_molecular_system
   _nef_molecular_system.sf_framecode  nef_molecular_system

   loop_
      _nef_sequence.index
      _nef_sequence.chain_code
      _nef_sequence.sequence_code
      _nef_sequence.residue_name
      _nef_sequence.linking
      _nef_sequence.residue_variant
      _nef_sequence.cis_peptide

     1   A    1   ALA   start    .   .
     2   A    2   ALA   middle   .   .
     3   A    3   ALA   end      .   .
     4   B   11   ALA   start    .   .
     5   B   12   ALA   middle   .   .
     6   B   13   ALA   end      .   .

   stop_

save_"""


# noinspection PyUnusedLocal
def test_3aa_x2_off_10_b():

    path = path_in_test_data(__file__, "3aa_x2.fasta")
    result = run_and_report(app, ["--starts", "1,11", path])

    mol_sys_result = isolate_frame(result.stdout, "%s" % NEF_MOLECULAR_SYSTEM)

    assert_lines_match(EXPECTED_3A_AB_B_start_11, mol_sys_result)


EXPECTED_3AA_HEADER_PARSING = """\
save_nef_molecular_system
   _nef_molecular_system.sf_category   nef_molecular_system
   _nef_molecular_system.sf_framecode  nef_molecular_system

   loop_
      _nef_sequence.index
      _nef_sequence.chain_code
      _nef_sequence.sequence_code
      _nef_sequence.residue_name
      _nef_sequence.linking
      _nef_sequence.residue_variant
      _nef_sequence.cis_peptide

     1   thx   -2   ALA   start    .   .
     2   thx   -1   ALA   middle   .   .
     3   thx    0   ALA   end      .   .

   stop_

save_"""


def test_3aa_header_parsing():

    path = path_in_test_data(__file__, "3aa_header.fasta")
    result = run_and_report(app, [path])

    mol_sys_result = isolate_frame(result.stdout, "%s" % NEF_MOLECULAR_SYSTEM)

    assert result.stdout.startswith("data_wibble")
    assert_lines_match(EXPECTED_3AA_HEADER_PARSING, mol_sys_result)


EXPECTED_3AA_X2_HEADER_PARSING = """\
save_nef_molecular_system
   _nef_molecular_system.sf_category   nef_molecular_system
   _nef_molecular_system.sf_framecode  nef_molecular_system

   loop_
      _nef_sequence.index
      _nef_sequence.chain_code
      _nef_sequence.sequence_code
      _nef_sequence.residue_name
      _nef_sequence.linking
      _nef_sequence.residue_variant
      _nef_sequence.cis_peptide

     1  X   -2   VAL   start    .   .
     2  X   -1   ILE   middle   .   .
     3  X    0   CYS   middle   .   .
     4  X    1   LYS   middle   .   .
     5  X    2   TYR   end      .   .
     6  Z   10   GLY   start    .   .
     7  Z   11   ALA   middle   .   .
     8  Z   12   ARG   middle   .   .
     9  Z   13   TYR   end      .   .

   stop_

save_"""


def test_3aa_x2_header_parsing():

    path = path_in_test_data(__file__, "3aa_x2_header.fasta")
    result = run_and_report(app, [path])

    assert Entry.from_string(result.stdout).entry_id == "test__test2"
    mol_sys_result = isolate_frame(result.stdout, "%s" % NEF_MOLECULAR_SYSTEM)

    assert_lines_match(EXPECTED_3AA_X2_HEADER_PARSING, mol_sys_result)


def test_3aa_x2_chain_repeat_header_parsing():

    file_name = "3aa_x2_header_repeat_chains.fasta"
    path = path_in_test_data(__file__, file_name)
    result = run_and_report(app, [path], expected_exit_code=1)

    assert "some of the input chain codes are repeated" in result.stdout
    assert "Z [2]" in result.stdout
    assert "all chain codes must be unique" in result.stdout
    assert file_name in result.stdout


@pytest.mark.parametrize(
    "header_or_text, expected_or_count",
    [
        (">pdb|4ciw|A", "pdb_4ciw_A"),
        (">CW42", "CW42"),
        (">crab_anapl ALPHA CRYSTALLIN B CHAIN (ALPHA(B)-CRYSTALLIN).", "crab_anapl"),
        (">gnl|mdb|bmrb16965:1 HET-s", "gnl_mdb_bmrb16965_1"),
        ("g", 1024),  # g repeated 1024 times
        (
            ">2DLV_1|Chain A|Regulator of G-protein signaling 18|Homo sapiens (9606)",
            "2DLV_1",
        ),
        (">2SRC_1|Chain A|TYROSINE-PROTEIN KINASE SRC|Homo sapiens (9606)", "2SRC_1"),
        (">H1N1 nucleoprotein, monomeric 416A mutant", "H1N1"),
        (">test NEFPLS | CHAIN: Z | START: 10", "test"),
        ("", "fasta"),
        (">", "fasta"),
    ],
)
def test_headers_2(header_or_text, expected_or_count, tmp_path):

    if is_int(expected_or_count):
        expected_or_count = header_or_text * expected_or_count
        header_or_text = f">{expected_or_count}"

    file_text = f"""
        {header_or_text}
        AAA
    """
    file_text = dedent(file_text)

    tmp_file = tmp_path / "test.fasta"
    with tmp_file.open("w") as fh:
        fh.write(file_text)

    result = run_and_report(app, [str(tmp_file)])

    entry = Entry.from_string(result.stdout)

    assert entry.entry_id == expected_or_count



import copy

import typer

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    read_test_data,
    run_and_report,
)
from nef_pipelines.tools.chains.align import align

app = typer.Typer()
app.command()(align)


def test_no_offset():

    INPUT = read_test_data("garyt.nef", __file__)

    result = run_and_report(app, [], input=INPUT)

    assert_lines_match(INPUT, result.stdout)


def test_offset_10():

    INPUT = read_test_data("garyt_offset_10.nef", __file__)
    EXPECTED = read_test_data("garyt.nef", __file__)

    result = run_and_report(app, [], input=INPUT)

    assert_lines_match(EXPECTED, result.stdout)


def test_offset_nearest_match():
    INPUT = read_test_data("garyt_offset_10_nearest.nef", __file__)
    EXPECTED = read_test_data("garyt.nef", __file__)
    EXPECTED = EXPECTED.replace("A   5       THR", "A   5       PHE")

    result = run_and_report(app, [], input=INPUT)

    assert_lines_match(result.stdout, EXPECTED)


def test_offset_mol_system_starts_m10():
    INPUT = read_test_data("garyt_offset_10_mol_system_starts_m10.nef", __file__)
    EXPECTED = copy.copy(INPUT)

    for i in range(10, 16):
        expected_fragment = f"A   {i}       "
        new_fragment = f"A   {i-21}       "
        EXPECTED = EXPECTED.replace(expected_fragment, new_fragment)

    result = run_and_report(app, [], input=INPUT)

    assert_lines_match(EXPECTED, result.stdout)


def test_offset_mol_system_starts_m10_is_superset():
    INPUT = read_test_data(
        "garyt_offset_10_mol_system_starts_m10_is_superset.nef", __file__
    )
    EXPECTED = copy.copy(INPUT)
    for i in range(10, 16):
        expected_fragment = f"A   {i}       "
        new_fragment = f"A   {i-21}       "
        EXPECTED = EXPECTED.replace(expected_fragment, new_fragment)

    result = run_and_report(app, [], input=INPUT)

    assert_lines_match(EXPECTED, result.stdout)


def test_multi_chain():
    INPUT = read_test_data("garyt_multi_chain.nef", __file__)
    EXPECTED = read_test_data("garyt_multi_chain_expected.nef", __file__)

    result = run_and_report(app, [], input=INPUT)

    assert_lines_match(EXPECTED, result.stdout)


def test_one_bad():

    INPUT = read_test_data("garyt_one_bad.nef", __file__)
    EXPECTED = read_test_data("garyt_one_bad_expected.nef", __file__)

    result = run_and_report(app, [], input=INPUT)

    assert_lines_match(EXPECTED, result.stdout)


def test_subset():

    INPUT = read_test_data("garyt_subset.nef", __file__)
    EXPECTED = read_test_data("garyt_subset_expected.nef", __file__)

    result = run_and_report(app, [], input=INPUT)

    assert_lines_match(EXPECTED, result.stdout)



from pathlib import Path

import typer

from nef_pipelines.lib.nef_lib import (
    molecular_system_from_entry_or_exit,
    read_entry_from_file_or_stdin_or_exit_error,
)
from nef_pipelines.lib.sequence_lib import chains_from_frames
from nef_pipelines.lib.util import STDIN, exit_error
from nef_pipelines.transcoders.fasta.exporters.sequence import pipe as fasta_pipe
from nef_pipelines.transcoders.mars import export_app


@export_app.command()
def sequence(
    chain_code: str = typer.Option(
        None,
        "-c",
        "--chain_code",
        help=" single chain to export  [default: 'A']",
        metavar="<CHAIN-CODE>",
    ),
    input_file: Path = typer.Option(
        STDIN, "-i", "--in", help="file to read nef data from", metavar="<NEF-FILE>"
    ),
    output_file: Path = typer.Argument(
        None,
        help="file name to output to [default <ENTRY-ID.fasta>] for stdout use -",
        metavar="<FASTA-SEQUENCE-FILE>",
    ),
    force: bool = typer.Option(
        False,
        "-f",
        "--force",
        help="force overwrite of output file if it exists and isn't empty",
    ),
):
    """- write a mars sequence file [fasta]"""

    entry = read_entry_from_file_or_stdin_or_exit_error(input_file)

    output_file = output_file if output_file else f"{entry.entry_id}.fasta"

    molecular_system = molecular_system_from_entry_or_exit(entry)

    chain_code = _get_single_chain_code_or_exit(
        chain_code, molecular_system, input_file
    )

    entry = fasta_pipe(entry, chain_code, Path(output_file), force)

    if entry:
        print(entry)


def _get_single_chain_code_or_exit(chain_code_selector, molecular_system, input_file):

    chain_codes = chains_from_frames(molecular_system)

    _exit_if_no_chain_codes(chain_codes, input)

    chain_code = None
    if not chain_code_selector:
        chain_code = _get_default_chain_code_or_exit(chain_codes, input_file)

    if not chain_code:
        chain_code = _select_single_chain_code_or_exit(chain_code_selector, chain_codes)

    return chain_code


def _exit_if_no_chain_codes(chain_codes, input):
    if len(chain_codes) != 1:
        msg = f"""\
            A  chain code is required and I didn't fina any in
            {input}
        """
        exit_error(msg)


def _select_single_chain_code_or_exit(chain_code, chain_codes):

    if chain_code not in chain_codes:
        chain_code_list = ", ".join(chain_codes)
        msg = f"""\
                the selected chain code is {chain_code} but this wasn't found in the input
                {input}
                this contained the following chain codes
                {chain_code_list}
            """
        exit_error(msg)

    return chain_code


def _get_default_chain_code_or_exit(chain_codes, input_file):

    if len(chain_codes) != 1:
        num_chains = len(chain_codes)
        chains_list = ", ".join(chain_codes)
        msg = f"""\
            You didn't select a chain code and single chain is required...
            I found {num_chains} chains in {input_file}
            the chains were:
            {chains_list}
        """
        exit_error(msg)

    return chain_codes[0]



from dataclasses import dataclass, field
from itertools import combinations
from math import sqrt
from statistics import stdev
from typing import Dict, List, OrderedDict, Tuple, Union

from fyeah import f
from ordered_set import OrderedSet
from pynmrstar import Entry, Loop, Saveframe
from streamfitter.interface import NoiseInfo

from nef_pipelines.lib.isotope_lib import (
    CODE_TO_ISOTOPE,
    GAMMA_RATIOS,
    MAGNETOGYRIC_RATIO_1H,
)
from nef_pipelines.lib.nef_lib import (
    NEF_RELAXATION_VERSION,
    UNUSED,
    SelectionType,
    create_nef_save_frame,
    loop_row_namespace_iter,
    select_frames,
)
from nef_pipelines.lib.peak_lib import frame_to_peaks
from nef_pipelines.lib.shift_lib import IntensityMeasurementType
from nef_pipelines.lib.structures import NEFPipelinesException

SERIES_DATA_CATEGORY = "_{NAMESPACE}_series_data"


class NEFPLSFitLibException(NEFPipelinesException):
    """Base exception class for NEF fit library errors."""

    ...


class NEFPLSFitLibNoIsotopeFrequenciesError(NEFPLSFitLibException):
    """Raised when no isotope frequencies are found in spectrum frames."""

    ...


class NEFPLSFitLibInconsistentSeriesDataError(NEFPLSFitLibException):
    """Raised when data is not consistent across series frames (frequencies or isotope axes)."""

    ...


class NEFPLSFitLibNoSeriesFramesSelectedError(NEFPLSFitLibException):
    """Raised when no series frames are selected by the given selectors."""

    ...


class NEFPLSFitLibNoFrameSelectorsError(NEFPLSFitLibException):
    """Raised when no frame selectors are provided."""

    ...


class NEFPLSFitLibMissingDataError(NEFPLSFitLibException):
    """Raised when expected data is missing (series data loop, spectrum frame, or data ID)."""

    ...


@dataclass
class RelaxationSeriesValues:
    spectra: List[str] = field(default_factory=list)
    peak_ids: List[int] = field(default_factory=list)
    variable_values: List[Union[int, float]] = field(default_factory=list)
    values: List[Union[int, float, bool]] = field(default_factory=list)


def _combine_relaxation_series(
    series1: RelaxationSeriesValues, series2: RelaxationSeriesValues
) -> RelaxationSeriesValues:
    """Combine two RelaxationSeriesValues into a single RelaxationSeriesValues.

    All fields from both series are combined as sequential blocks (series1 data first,
    then series2 data). Raises an exception if equivalent fields between the two series
    have different lengths.

    Args:
        series1: First RelaxationSeriesValues to combine
        series2: Second RelaxationSeriesValues to combine

    Returns:
        RelaxationSeriesValues with combined data

    Raises:
        NEFPLSFitLibInconsistentSeriesDataError: If equivalent fields have different lengths between series
    """
    # Validate that equivalent fields between series1 and series2 have the same length
    field_comparisons = [
        ("spectra", len(series1.spectra), len(series2.spectra)),
        ("peak_ids", len(series1.peak_ids), len(series2.peak_ids)),
        ("variable_values", len(series1.variable_values), len(series2.variable_values)),
        ("values", len(series1.values), len(series2.values)),
    ]

    for field_name, len1, len2 in field_comparisons:
        if len1 != len2:
            msg = f"Inconsistent field lengths for {field_name}: series1 has {len1}, series2 has {len2}"
            raise NEFPLSFitLibInconsistentSeriesDataError(msg)

    # Combine all fields as blocks (series1 first, then series2)
    combined = RelaxationSeriesValues(
        spectra=[*series1.spectra, *series2.spectra],
        peak_ids=[*series1.peak_ids, *series2.peak_ids],
        variable_values=[*series1.variable_values, *series2.variable_values],
        values=[*series1.values, *series2.values],
    )

    return combined


def _fit_results_as_frame(
    series_frame,
    prefix,
    entry,
    fits,
    fit_name,
    output,
    monte_carlo_errors,
    monte_carlo_value_stats,
    monte_carlo_param_values,
    monte_carlo_random_seed,
    noise_info: NoiseInfo,
    version_strings,
    fitter_name,
):

    spectrum_frames = _series_frame_to_spectrum_frames(series_frame, prefix, entry)

    data_id_to_spectrum_peak_ids = _series_frame_to_data_id_to_spectrum_peak_id(
        series_frame, prefix
    )

    spectrum_peak_id_to_axis_atoms = _build_spectrum_peak_id_to_axis_atoms(
        spectrum_frames
    )

    isotope_axes = spectra_to_isotope_axes(spectrum_frames)

    _exit_if_isotope_axes_are_not_unique(isotope_axes, spectrum_frames, entry.entry_id)

    isotope_frequencies = spectra_to_isotope_frequencies(spectrum_frames)

    _exit_if_isotope_frequencies_are_not_unique(
        isotope_frequencies, spectrum_frames, entry.entry_id
    )

    isotope_frequencies = _get_unique_isotope_frequencies(isotope_frequencies)

    _exit_if_no_frequencies_found(isotope_frequencies, series_frame, entry)

    field_strength = frequencies_to_field_strength(isotope_frequencies)

    axis_of_lowest_gamma = _isotope_axes_to_lowest_gamma_axis(isotope_axes)

    result_frame = create_nef_save_frame(f"{prefix}_relaxation_list", output)

    result_frame.add_tag("version", NEF_RELAXATION_VERSION)
    result_frame.add_tag("experiment_type", series_frame.get_tag("experiment_type")[0])
    result_frame.add_tag("field_strength", field_strength)
    result_frame.add_tag("value_unit", "s-1")
    result_frame.add_tag("relaxation_atom_id", axis_of_lowest_gamma)
    result_frame.add_tag("ref_value", UNUSED)
    result_frame.add_tag("source", "experimental")
    result_frame.add_tag("fitting_function", fitter_name)
    # TODO fix this
    result_frame.add_tag("minimizer", "leastsq")

    result_frame.add_tag("error_method", noise_info.source)

    # TODO move these to tags under nefpls...
    comment = f"""
        fitting software: {version_strings}
        random seed {monte_carlo_random_seed}
        requested noise estimate source: {noise_info.requested_noise_source}
        source of noise estimate: {noise_info.source}
        noise estimate: {noise_info.noise}
        noise estimate fractional error: {noise_info.fraction_error_in_noise}
        noise estimate number replicates: {noise_info.num_replicates}
    """
    # error_in_noise-estimate {}
    # fit_time {}
    result_frame.add_tag("comment", comment)

    relaxation_loop = Loop.from_scratch(f"{prefix}_relaxation")
    result_frame.add_loop(relaxation_loop)

    RELAXATION_LOOP_TAGS = "index data_id data_combination_id value value_error".split()
    RELAXATION_LOOP_ATOM_TAGS = (
        "chain_code_{axis} sequence_code_{axis} residue_name_{axis} atom_name_{axis}"
    )

    all_tags = [
        *RELAXATION_LOOP_TAGS,
    ]
    for axis in range(1, len(isotope_axes) + 1):
        all_tags.append(RELAXATION_LOOP_ATOM_TAGS.format(axis=axis).split())

    relaxation_loop.add_tag(all_tags)

    data = []
    for index, (data_id, fit) in enumerate(fits.items(), start=1):
        data_row = {"index": index, "data_id": index, "data_combination_id": UNUSED}

        peak_atoms = []
        atom_key = next(iter(data_id_to_spectrum_peak_ids[data_id]))

        # TODO check if peak exists or error
        for axis_atoms in spectrum_peak_id_to_axis_atoms[atom_key].values():
            peak_atoms.extend(axis_atoms)

        for i, atom in enumerate(peak_atoms, start=1):
            residue = atom.residue
            data_row.update(
                {
                    f"chain_code_{i}": residue.chain_code,
                    f"sequence_code_{i}": residue.sequence_code,
                    f"residue_name_{i}": residue.residue_name,
                    f"atom_name_{i}": atom.atom_name,
                }
            )

        if monte_carlo_errors:
            mc_error = monte_carlo_errors[data_id].get(f"{fit_name}_mc_error", UNUSED)
        else:
            mc_error = UNUSED
        data_row.update(
            {
                "value": f"{fit.params[fit_name].value:.6f}",
                "value_error": f"{mc_error:.6}",
            }
        )
        data.append(data_row)

    relaxation_loop.add_data(data)

    return result_frame


def _series_frame_to_data_id_to_spectrum_peak_id(series_frame, prefix):
    result = {}
    series_loop_id = f"_{prefix}_series_data"

    series_loop = (
        series_frame[series_loop_id] if series_loop_id in series_frame else None
    )
    for row in loop_row_namespace_iter(series_loop):
        spectrum_peak_id = row.nmr_spectrum_id, row.peak_id
        result.setdefault(row.data_id, set()).add(spectrum_peak_id)
    return result


def _build_spectrum_peak_id_to_axis_atoms(spectrum_frames):
    result = {}
    for spectrum_frame in spectrum_frames:
        peaks = frame_to_peaks(spectrum_frame)

        for peak in peaks:
            axis_atoms = {}
            key = spectrum_frame.name, peak.id
            result[key] = axis_atoms

            for i, shift in enumerate(peak.shifts, start=1):
                axis_atoms.setdefault(i, set()).add(shift.atom)

    return result


def _isotope_axes_to_lowest_gamma_axis(isotope_axes):
    return min(
        isotope_axes,
        key=lambda x: min(GAMMA_RATIOS[isotope] for isotope in isotope_axes[x]),
    )


def _exit_if_no_frequencies_found(isotope_frequencies, series_frame, entry):
    if not isotope_frequencies:
        msg = f"""
            no isotope frequencies found in the series frame {series_frame.name} in entry {entry.entry_id}
            this indcates that the series frame is incomplete or missing...
        """
        raise NEFPLSFitLibNoIsotopeFrequenciesError(msg)


def _get_unique_isotope_frequencies(isotope_frequencies):

    return {
        isotope: list(frequency)[0]
        for isotope, frequency in isotope_frequencies.items()
    }


def _series_frame_to_spectrum_frames(series_frame, prefix, entry):

    spectrum_frames = []

    for i, row in enumerate(
        loop_row_namespace_iter(
            series_frame.get_loop(f"{prefix}_series_experiment"), convert=True
        ),
        start=1,
    ):
        spectrum_frame_name = row.nmr_spectrum_id
        if spectrum_frame_name in entry.frame_dict.keys():
            spectrum_frame = entry.get_saveframe_by_name(spectrum_frame_name)
            spectrum_frames.append(spectrum_frame)

    return spectrum_frames


# TODO:  there is some duplication here!
def _select_relaxation_series_or_exit(
    entry: Entry, frame_selectors: List[str]
) -> Tuple:
    relaxation_frames = select_frames(
        entry, "nefpls_series_list", SelectionType.CATEGORY
    )
    series_frames_and_ids = OrderedSet(
        [
            (relaxation_series.name, id(relaxation_series))
            for relaxation_series in relaxation_frames
        ]
    )

    if frame_selectors:
        named_frames = []
        for frame_selector in frame_selectors:
            named_frames.extend(
                select_frames(entry, frame_selector, SelectionType.NAME)
            )

        named_frames_and_ids = OrderedSet(
            [(named_frame.name, id(named_frame)) for named_frame in named_frames]
        )
    elif len(relaxation_frames) == 1:
        relaxation_frame = relaxation_frames[0]
        frame_key = (relaxation_frame.name, id(relaxation_frame))
        named_frames_and_ids = OrderedSet(
            [
                frame_key,
            ]
        )
    else:
        # No frame selectors provided and multiple frames available - use all relaxation frames
        named_frames_and_ids = series_frames_and_ids

    selected_frames_and_ids = series_frames_and_ids.intersection(named_frames_and_ids)

    result = [
        entry.get_saveframe_by_name(frame_name)
        for frame_name, _ in selected_frames_and_ids
    ]

    return result


def _exit_if_no_series_frames_selected(active_relaxation_frames, frame_selectors):
    if not active_relaxation_frames:
        msg = f"""
            no series frames selected by the selectors
            {" ".join(frame_selectors)}
        """
        raise NEFPLSFitLibNoSeriesFramesSelectedError(msg)


def _exit_if_no_frame_selectors(frame_selectors):
    if len(frame_selectors) == 0:
        msg = "you must select some frames!"
        raise NEFPLSFitLibNoFrameSelectorsError(msg)

    return None


def _exit_if_isotope_frequencies_are_not_unique(
    isotope_frequencies, series_frame, entry_id
):
    # TODO: should identify common frames and swapped ones
    for isotope, frequencies in isotope_frequencies.items():
        if len(frequencies) > 1:
            msg = f"""
                in the series frame {series_frame.name} in entry {entry_id}
                isotope {isotope} has multiple frequencies {frequencies}
                the frequencies are {', '.join(frequencies)}
                spectrometer frequencies must be consistent in a series frame
            """

            raise NEFPLSFitLibInconsistentSeriesDataError(msg)


def _exit_if_isotope_axes_are_not_unique(isotope_axes, series_frame, entry_id):

    # TODO: should identify common frames and swapped ones
    for axis, isotopes in isotope_axes.items():
        if len(isotopes) > 1:
            msg = f"""
                in the series frame {series_frame.name} in entry {entry_id}
                the isotopes {', '.join(isotopes)} appears on multiple different axes
                this variation is not compatible with the nef relaxation lists model
                the axes must be consistent in a series frame
            """

            raise NEFPLSFitLibInconsistentSeriesDataError(msg)


def spectra_to_isotope_axes(spectrum_frames):
    isotope_axes = {}
    for spectrum_frame in spectrum_frames:
        spectrum_dimension_loop = loop_row_namespace_iter(
            spectrum_frame.get_loop("nef_spectrum_dimension")
        )

        for row in spectrum_dimension_loop:
            isotope = CODE_TO_ISOTOPE.get(row.axis_code)
            isotope_axes.setdefault(row.dimension_id, set()).add(isotope)

    return isotope_axes


def spectra_to_isotope_frequencies(spectrum_frames):

    frequencies = {}
    for spectrum_frame in spectrum_frames:
        spectrum_dimension_loop = loop_row_namespace_iter(
            spectrum_frame.get_loop("nef_spectrum_dimension")
        )
        for row in spectrum_dimension_loop:
            frequencies.setdefault(row.axis_code, set()).add(row.spectrometer_frequency)

    return frequencies


def frequencies_to_field_strength(frequencies):
    fields_by_gamma_and_isotope = {}
    for isotope_code, frequency in frequencies.items():

        if isotope_code not in CODE_TO_ISOTOPE:
            continue

        isotope = CODE_TO_ISOTOPE[isotope_code]
        gamma_ratio = GAMMA_RATIOS[isotope]

        field = frequency / MAGNETOGYRIC_RATIO_1H / gamma_ratio

        fields_by_gamma_and_isotope[gamma_ratio, isotope] = field

    highest_gamma_isotope = max(fields_by_gamma_and_isotope, key=lambda x: x[0])

    return fields_by_gamma_and_isotope[highest_gamma_isotope]


def _get_spectra_by_series_variable(entry, series_experiment_loop):
    from nef_pipelines.lib.nef_lib import loop_row_namespace_iter

    spectra_by_times = {}
    for i, row in enumerate(
        loop_row_namespace_iter(series_experiment_loop, convert=True), start=1
    ):
        spectrum_frame_name = row.nmr_spectrum_id
        series_variable = row.series_variable

        spectrum_frame = entry.get_saveframe_by_name(spectrum_frame_name)

        key = i, series_variable, spectrum_frame_name
        spectra_by_times[key] = spectrum_frame

    return spectra_by_times


def _exit_if_spectra_are_missing(spectra_by_times_and_indices, series_name):
    for (
        _,
        _,
        spectrum_frame_name,
    ), spectrum_frame in spectra_by_times_and_indices.items():
        if not spectrum_frame:
            msg = f"no spectrum frame found for series {series_name} for spectrum {spectrum_frame_name}"
            raise NEFPLSFitLibMissingDataError(msg)


def _get_atoms_and_values(
    peaks_by_times_and_indices, data_type: IntensityMeasurementType
) -> Dict:
    atoms_to_values = {}
    for (index, x_value, spectrum_name), peaks in peaks_by_times_and_indices.items():

        for peak in peaks:
            peak_atoms = tuple([shift.atom for shift in peak.shifts])
            x_and_y = atoms_to_values.setdefault(
                tuple(peak_atoms), RelaxationSeriesValues()
            )
            x_and_y.peak_ids.append(peak.id)
            x_and_y.variable_values.append(x_value)
            y_value = (
                peak.height
                if data_type == IntensityMeasurementType.HEIGHT
                else peak.volume
            )
            x_and_y.values.append(y_value)
            x_and_y.spectra.append(spectrum_name)

    return atoms_to_values


def calculate_noise_level_from_replicates(xy_data: RelaxationSeriesValues) -> float:

    repetitons = []

    for series_values in xy_data.values():
        xs = series_values.variable_values
        ys = series_values.values
        ys_by_x = {}
        for x, y in zip(xs, ys):
            ys_by_x.setdefault(float(x), []).append(float(y))
        repeated_values = [values for values in ys_by_x.values() if len(values) > 1]
        repetitons.extend(repeated_values)

    differences = []
    for repetiton_set in repetitons:
        for combination in combinations(repetiton_set, 2):
            differences.append(combination[0] - combination[1])

    noise_level = stdev(differences) if differences else None

    fraction_error_in_stdev = (
        1 / sqrt(2) * 1 / sqrt(len(differences)) if differences else None
    )

    return noise_level, fraction_error_in_stdev, len(differences)


def _series_frame_to_outputs(series_frame: Saveframe, prefix: str, entry: Entry):
    NAMESPACE = prefix  # noqa: F841
    series_category = f(SERIES_DATA_CATEGORY)
    series_data_loop = (
        series_frame.get_loop(series_category)
        if series_category in series_frame
        else None
    )

    _exit_if_no_series_data_loop(series_data_loop, series_frame, entry, series_category)

    ids_to_outputs = OrderedDict()
    for row in loop_row_namespace_iter(series_data_loop):

        data_id = row.data_id
        if row.relaxation_list_id != UNUSED:
            ids_to_outputs.setdefault(data_id, OrderedSet()).add(row.relaxation_list_id)

    return ids_to_outputs


def _series_frame_to_id_series_data(series_frame: Saveframe, prefix: str, entry: Entry):

    NAMESPACE = prefix  # noqa: F841
    series_category = f(SERIES_DATA_CATEGORY)
    series_data_loop = (
        series_frame.get_loop(series_category)
        if series_category in series_frame
        else None
    )

    _exit_if_no_series_data_loop(series_data_loop, series_frame, entry, series_category)

    de_duped_data = OrderedDict()
    for row in loop_row_namespace_iter(series_data_loop):
        data_id = row.data_id

        nmr_spectrum_id = row.nmr_spectrum_id
        peak_id = row.peak_id
        variable_value = row.variable_value
        value = row.value

        values = (nmr_spectrum_id, peak_id, variable_value, value)

        de_duped_data.setdefault(data_id, OrderedSet()).add(values)

    id_to_xy_data = {}
    for data_id, rows in de_duped_data.items():
        relaxation_series = id_to_xy_data.setdefault(data_id, RelaxationSeriesValues())
        for row in rows:
            nmr_spectrum_id, peak_id, variable_value, value = row
            relaxation_series.spectra.append(nmr_spectrum_id)
            relaxation_series.peak_ids.append(peak_id)
            relaxation_series.variable_values.append(variable_value)
            relaxation_series.values.append(value)

    return id_to_xy_data


def _exit_if_no_series_data_loop(
    series_data_loop, series_frame, entry, series_category
):
    if series_data_loop is None:
        msg = f"""
            in the entry {entry.entry_id}
            the series frame {series_frame.name} does not have a {series_category} loop
        """
        raise NEFPLSFitLibMissingDataError(msg)



import os
import string
from collections import Counter
from dataclasses import replace
from enum import auto
from pathlib import Path
from textwrap import dedent
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

from ordered_set import OrderedSet
from pynmrstar import Entry, Loop, Saveframe
from strenum import LowercaseStrEnum

from nef_pipelines.lib.constants import NEF_UNKNOWN
from nef_pipelines.lib.nef_lib import UNUSED, loop_row_namespace_iter

# from nef_pipelines.lib.nef_lib import loop_to_dataframe
from nef_pipelines.lib.structures import AtomLabel, Linking, SequenceResidue
from nef_pipelines.lib.util import (
    chunks,
    exit_error,
    get_display_file_name,
    is_int,
    read_from_file_or_exit,
    strip_characters_right,
)


class MoleculeType(LowercaseStrEnum):
    PROTEIN = auto()
    RNA = auto()
    DNA = auto()
    CARBOHYDRATE = auto()


NEF_CHAIN_CODE = "chain_code"
ANY_CHAIN = "_"


class ChainCodeIterable:

    ASCII_UPPERCASE = list(string.ascii_uppercase)

    def __init__(self, user_chain_codes: List[str], exclude: List[str] = ()):

        self._index = 0
        self._chain_codes = []

        seen_codes = set()
        self._chain_codes = []
        for user_chain_code in user_chain_codes:
            seen_codes.add(user_chain_code)
            if user_chain_code not in exclude:
                self._chain_codes.append(user_chain_code)

        for chain_code in self.ASCII_UPPERCASE:
            if chain_code not in seen_codes and chain_code not in exclude:
                self._chain_codes.append(chain_code)

    def __iter__(self):
        return self

    def __next__(self):
        if self._index < len(self._chain_codes):
            result = self._chain_codes[self._index]
            self._index += 1
            return result
        else:
            raise StopIteration("run out of chain names!")


def get_chain_code_iter(
    user_chain_codes: List[str] = "", exclude: List[str] = ()
) -> Iterator[str]:
    """
    return an iterator they will yield chain codes in user chain codes and
    once exhausted yield any remaining letters of the upper case alphabet
    till they run out. A list of chain codes to not use can be provided...

    Args:
        user_chain_codes (str):  string of dot separated chain codes or a list of chain codes
        exclude: chain codes to not include in the iteration

    Returns:
        Iterable[str] single codes
    """

    return ChainCodeIterable(user_chain_codes, exclude).__iter__()


def get_linking(
    target_sequence: List[SequenceResidue],
    no_chain_start=List[str],
    no_chain_end=List[str],
) -> List[SequenceResidue]:
    """
    get the correct linking for residue in a standard sequenced chain

    Args:
        target_sequence (List[SequenceResidue]) : the sequence
        no_chain_start (List[str]): if the chain isn't in the list cap it at the start
        no_end (List[str]): if the chain isn't in the list cap it at the end

    Returns:
         List[SequenceResidue]: correctly linked residues
    """

    by_chain = {}
    for residue in target_sequence:
        by_chain.setdefault(residue.chain_code, []).append((residue, "middle"))

    result = []
    for chain, residues_and_linkages in by_chain.items():

        if chain not in no_chain_start:
            residues_and_linkages[0] = (residues_and_linkages[0][0], "start")

        if chain not in no_chain_end:
            residues_and_linkages[-1] = (residues_and_linkages[-1][0], "end")

        result.extend(residues_and_linkages)

    return result


def sequence_to_nef_frame(
    sequences: List[SequenceResidue],
    no_chain_start: Union[List[str], Tuple[str]] = (),
    no_chain_end: Union[List[str], Tuple[str]] = (),
) -> Saveframe:
    """

    Args:
        sequences (List[SequenceResidue]): the sequence as a list of sequence residues [will be sorted]
        no_chain_start (List[str]): if the chain isn't in the list cap it at the start
        no_chain_end (List[str]): if the chain isn't in the list cap it at the end

    Returns:
        Saveframe: a NEF saveframe
    """

    category = "nef_molecular_system"

    frame_code = f"{category}"

    nef_frame = Saveframe.from_scratch(frame_code, category)

    nef_frame.add_tag("sf_category", category)
    nef_frame.add_tag("sf_framecode", frame_code)

    nef_loop = Loop.from_scratch("nef_sequence")
    nef_frame.add_loop(nef_loop)

    tags = (
        "index",
        "chain_code",
        "sequence_code",
        "residue_name",
        "linking",
        "residue_variant",
        "cis_peptide",
    )

    nef_loop.add_tag(tags)

    # TODO need tool to set ionisation correctly
    residue_and_linkages = get_linking(sequences, no_chain_start, no_chain_end)

    for index, (sequence_residue, linking) in enumerate(residue_and_linkages):

        data = [
            {
                "index": index + 1,
                "chain_code": sequence_residue.chain_code,
                "sequence_code": sequence_residue.sequence_code,
                "residue_name": sequence_residue.residue_name.upper(),
                "linking": linking,
                "residue_variant": NEF_UNKNOWN,
                "cis_peptide": NEF_UNKNOWN,
            }
        ]
        nef_loop.add_data(data)

    return nef_frame


class MoleculeTypes(LowercaseStrEnum):
    PROTEIN = auto()
    DNA = auto()
    RNA = auto()
    CARBOH = auto()
    LIGAND = auto()
    OTHER = auto()


TRANSLATIONS_3_1_DNA = {"DG": "G", "DC": "C", "DT": "T", "DA": "A"}
TRANSLATIONS_1_3_DNA = {value: key for (key, value) in TRANSLATIONS_3_1_DNA.items()}

TRANSLATIONS_3_1_RNA = {"G": "G", "C": "C", "A": "A", "U": "U", "I": "I"}
TRANSLATIONS_1_3_RNA = {value: key for (key, value) in TRANSLATIONS_3_1_RNA.items()}

TRANSLATIONS_3_1_PROTEIN = {
    "ALA": "A",
    "ARG": "R",
    "ASN": "N",
    "ASP": "D",
    "CYS": "C",
    "GLU": "E",
    "GLN": "Q",
    "GLY": "G",
    "HIS": "H",
    "ILE": "I",
    "LEU": "L",
    "LYS": "K",
    "MET": "M",
    "PHE": "F",
    "PRO": "P",
    "SER": "S",
    "THR": "T",
    "TRP": "W",
    "TYR": "Y",
    "VAL": "V",
}
TRANSLATIONS_1_3_PROTEIN = {
    value: key for (key, value) in TRANSLATIONS_3_1_PROTEIN.items()
}

TRANSLATIONS_1_3 = {
    MoleculeTypes.PROTEIN: TRANSLATIONS_1_3_PROTEIN,
    MoleculeTypes.DNA: TRANSLATIONS_1_3_DNA,
    MoleculeTypes.RNA: TRANSLATIONS_1_3_RNA,
    MoleculeTypes.CARBOH: None,
    MoleculeTypes.OTHER: None,
    MoleculeTypes.LIGAND: None,
}
TRANSLATIONS_3_1 = {
    MoleculeTypes.PROTEIN: TRANSLATIONS_3_1_PROTEIN,
    MoleculeTypes.DNA: TRANSLATIONS_3_1_DNA,
    MoleculeTypes.RNA: TRANSLATIONS_3_1_RNA,
    MoleculeTypes.CARBOH: None,
    MoleculeTypes.OTHER: None,
    MoleculeTypes.LIGAND: None,
}


class BadResidue(Exception):
    """
    Bad residue found in a sequence
    """

    def __init__(self, residue_name, error_pos, sequence_string):

        self.residue_name = residue_name
        self.sequence_string = sequence_string
        self.error_pos = error_pos

    def __str__(self):
        try:
            terminal_width = os.get_terminal_size().columns
        except OSError:
            terminal_width = 80

        error_pos_string = [
            " ",
        ] * len(self.sequence_string)
        error_pos_string[self.error_pos] = "^"
        error_pos_string = "".join(error_pos_string)

        msg = f"""\
        unknown residue {self.residue_name}
        at residue {self.error_pos+1}
        sequence:
        """
        msg = dedent(msg)
        sequence_string_chunks = list(chunks(self.sequence_string, terminal_width - 10))
        pos_string_chunks = list(chunks(error_pos_string, terminal_width - 10))
        for sequence_chunk, pos_chunk in zip(sequence_string_chunks, pos_string_chunks):
            msg += f"{''.join(sequence_chunk)}\n"
            msg += f"{''.join(pos_chunk)}\n"

        return msg


def translate_1_to_3(
    sequence: str, molecule_type=MoleculeTypes.PROTEIN, unknown: Optional[str] = None
) -> List[str]:
    """
    Translate a 1 letter sequence to a 3 letter sequence
    Args:
        sequence (str): 1 letter sequence
        molecule_type (MoleculeTypes): type of molecule to translate residue types
        unknown (Optional[str]): optional name for residues if they are unknown, if set no error is raised if a
                                 1 letter residue name is not recognised
    Returns List[str]:
        a list of 3 residue codes
        :param type:

    """

    translations = TRANSLATIONS_1_3[molecule_type]

    result = []
    for i, residue_name_1let in enumerate(sequence):
        if residue_name_1let == " ":
            continue
        original_residue_name_1let = residue_name_1let
        residue_name_1let = residue_name_1let.upper()
        if translations is None:
            result.append(residue_name_1let)
        elif residue_name_1let in translations:
            result.append(translations[residue_name_1let])
        else:
            if unknown:
                result.append(unknown)
            else:
                raise BadResidue(original_residue_name_1let, i, sequence)

    return result


def translate_3_to_1(
    sequence: Union[str, List[str]], molecule_type=MoleculeTypes.PROTEIN
) -> List[str]:
    """

    Translate a 3 letter sequence to a 1 letter sequence. If a single string is supplied it
    is trabnslated and returned as a sinle residue
    Args:
        sequence (str|List[str]): 3 letter sequence code or 3 letter sequence
        molecule_type (MoleculeTypes): type of molecule to translate residue types

    Returns str|List[str]:
        a list of 1 residue codes or a single 1 letter residue code

    :param sequence:
    :return:
    """

    single_aa = False
    if isinstance(sequence, str):
        sequence = [
            str,
        ]
        single_aa = True

    translations = TRANSLATIONS_3_1[molecule_type]

    result = []
    for i, residue_name in enumerate(sequence, start=1):
        residue_name = residue_name.upper()
        if translations is None:
            if len(residue_name) == 1:
                result.append(residue_name)
            else:
                msg = f"""
                    it isn't possible to translate the residue name {residue_name} to a 1 letter code
                    the molecule type {molecule_type} doesn't have one letter codes and the residue name: {residue_name}
                    is longer than one letter and so can't be passed through
                """
                exit_error(msg)
        if residue_name in translations:
            result.append(translations[residue_name])
        else:
            msg = f"""
            unknown residue {residue_name}
            in sequence {chunks(' '.join(sequence), 10)}
            at residue number {i}
            """
            exit_error(msg)

    if single_aa:
        result = result[0]

    return result


def sequence_3let_to_sequence_residues(
    sequence_3let: List[str], chain_code: str = "A"
) -> List[SequenceResidue]:
    """
    Translate a list of 3 residue sequence codes to SequenceResidues
    Args:
        sequence_3let (List([str]): list of 3 letter residue names
        chain_code (str): the chain code [defaults to A]
        offset (int): the sequence offset [defaults to 0]

    Returns List[SequenceResidue]:
        the sequence as a list of SequenceResidues

    """

    return [
        SequenceResidue(chain_code, i, residue)
        for (i, residue) in enumerate(sequence_3let, start=1)
    ]


def sequence_residues_to_sequence_3let(
    sequence: List[SequenceResidue], chain_code: str = "A"
) -> List[str]:
    """
    Translate a list of SequenceResidues to 3 letter sequence codes t
    Args:
        sequence (List([SequenceResidues]): list of residues
        chain_code (str): the chain code [defaults to A] only residues from this chain will be converted

    Returns List[str]:
        the sequence as a list of 3 letter residues

    """
    return [
        residue.residue_name for residue in sequence if residue.chain_code == chain_code
    ]


def chains_from_frames(
    sequence_frame_or_frames: Union[Saveframe, List[Saveframe]]
) -> List[str]:
    """
     a nef molecular systems list the chains found

    :param sequence_frames: a nef molecular system save frame
    :return: a list of chain_codes
    """
    if isinstance(sequence_frame_or_frames, Saveframe):
        sequence_frames = [
            sequence_frame_or_frames,
        ]
    else:
        sequence_frames = sequence_frame_or_frames

    chains = set()
    for sequence_frame in sequence_frames:
        sequences = sequences_from_frames(sequence_frame)
        for residue in sequences:
            chains.add(residue.chain_code)

    if NEF_UNKNOWN in chains:
        chains.remove(NEF_UNKNOWN)

    if None in chains:
        chains.remove(None)

    chains = sorted(chains)

    return chains


def count_residues(sequence_frame: Saveframe, chain_code: str) -> Dict[str, int]:
    """
    given a nef molecular system and a chain list the number of residues of each type in the chain

    :param sequence_frames: a nef molecular system save frame
    :return: a list of chain_codes
    """

    residue_number_to_residue_name = {}
    for row in loop_row_namespace_iter(sequence_frame.loops[0]):
        if row.chain_code == chain_code:
            residue_number_to_residue_name[row.sequence_code] = row.residue_name

    result = Counter()
    for residue_name in sorted(residue_number_to_residue_name.values()):
        result[residue_name] += 1

    return result


def get_sequence_or_exit(
    file_name: Path = Path("-"),
) -> Optional[List[SequenceResidue]]:
    """
    read a sequence from a nef molecular system on a file or stdin or exit, return a list of residues

    file_name: a path to read the nef from Path('-') read from stdin
    :return: a list of parsed residues and chains, in the order they were read
    """
    try:
        result = get_sequence(file_name)
    except Exception as e:
        msg = "couldn't read sequence from nef input stream, either no stream or no nef molecular system frame"
        exit_error(msg, e)

    if len(result) == 0:
        msg = "no sequence read from nef molecular system in nef input stream"
        exit_error(msg)

    return result


def get_sequence(file_name: Path = Path("-")) -> List[SequenceResidue]:
    """
    read a sequence from a nef molecular system on stdin and return a list of residues

    file_name: a path to read the nef from Path('-') read from stdin
    can raise Exceptions

    :return:a list of parsed residues and chains, in the order they were read
    """

    text = read_from_file_or_exit(file_name, "sequence from NEF molecular system")

    display_file_name = get_display_file_name(file_name)

    try:
        entry = Entry.from_string(text)
    except Exception as e:
        msg = f"couldn't parse text in {display_file_name} as NEF data"
        exit_error(msg, e)

    return sequence_from_entry(entry)


def sequence_from_entry_or_exit(entry: Entry) -> List[SequenceResidue]:
    sequence = sequence_from_entry(entry)

    if not sequence:
        msg = "did you read a sequence? a sequence is required and was not read"
        exit_error(msg)

    return sequence


def sequence_from_entry(entry) -> List[SequenceResidue]:

    result = []
    if entry is not None:
        frames = entry.get_saveframes_by_category("nef_molecular_system")

        if frames:
            result = sequences_from_frames(frames[0])
    return result


# TODO: check how this deals with sequences with repeats e.g. a sequence with A.3.ALA amd A.3.GLY etc
def sequences_from_frames(
    frame_or_frames: Union[Saveframe, List[Saveframe]],
    chain_codes_to_select: Union[str, List[str]] = ANY_CHAIN,
) -> List[SequenceResidue]:
    """
    read sequences from a nef save frame with the maximal information content about the residues

    :param frame: the save frame to read residues from, must have category nef molecular system
    :param chain_code_to_select: the chain codes to select this can be either a string or list of strings,
                                 any chain is selected using the instance of the constant string ANY_CHAIN [default]
    :return: a list of parsed residues and chains, in the order they were read
    """

    # TODO: check chain code for validity
    if chain_codes_to_select is not ANY_CHAIN and isinstance(
        chain_codes_to_select, str
    ):
        chain_codes_to_select = [
            chain_codes_to_select,
        ]

    if chain_codes_to_select is not ANY_CHAIN and isinstance(
        chain_codes_to_select, str
    ):
        chain_codes_to_select = set(
            [
                chain_codes_to_select,
            ]
        )
    elif chain_codes_to_select is not ANY_CHAIN:
        chain_codes_to_select = set(chain_codes_to_select)

    if isinstance(frame_or_frames, Saveframe):
        frame_or_frames = [
            frame_or_frames,
        ]

    residues = OrderedSet()
    for frame in frame_or_frames:
        for loop in frame.loops:
            residues.update(_parse_loops_residues(loop, chain_codes_to_select))

    return select_best_residues_by_info_content(residues)


def select_best_residues_by_info_content(residues):
    """
    given a list of residues select the best residue definition for each chain code and sequence code
    this removes formally unassigned residues [contain # or @ in sequence_code or chain_code] and favours a residue
    with a residue name, the priority is then cis peptide > residue variants > linking though this shouldn't have an
    effect unless there are errors in the definition of a molecular system

    note if there are multiple residues with the same residue name they are both added...

    :param residues: a list of residues

    :return: a list of residues with the best description possible
    """
    residues = _filter_partially_or_unassigned_residues(residues)

    residues_by_chain_code_sequence_code = {}
    for residue in residues:
        key = residue.chain_code, residue.sequence_code
        residues_by_chain_code_sequence_code.setdefault(key, []).append(residue)

    new_residues = []
    for i, equivalent_residues in enumerate(
        residues_by_chain_code_sequence_code.values()
    ):

        residues_and_scores = {}
        score = 0
        for equivalent_residue in equivalent_residues:
            # somewhat arbitrary scoring system to pick the best residue
            # TODO should we merge if there is missing info...
            if equivalent_residue.residue_name not in (None, UNUSED):
                score += 100000
            if equivalent_residue.is_cis not in (None, UNUSED):
                score += 10000
            if equivalent_residue.variants not in (None, UNUSED):
                score += 1000
            if equivalent_residue.linking not in (None, UNUSED):
                score += 100
            residues_and_scores[score] = equivalent_residue

        equivalent_residues = [
            residues_and_scores[score] for score in sorted(residues_and_scores)
        ]

        new_equivalent_residues = []
        if len(equivalent_residues) > 1:

            seen_residue_names = set()
            for equivalent_residue in equivalent_residues:
                if equivalent_residue.residue_name not in (None, UNUSED):
                    if equivalent_residue.residue_name not in seen_residue_names:
                        new_equivalent_residues.append(equivalent_residue)
                        seen_residue_names.add(equivalent_residue.residue_name)

            if new_equivalent_residues:
                new_equivalent_residues = [
                    equivalent_residues[0],
                ]
        else:
            new_equivalent_residues = [
                equivalent_residues[0],
            ]

        new_residues.extend(new_equivalent_residues)
    return new_residues


def _filter_partially_or_unassigned_residues(residues):

    result = []
    for residue in residues:
        if residue.chain_code in (None, UNUSED):
            continue
        if residue.sequence_code in (None, UNUSED):
            continue
        if isinstance(residue.chain_code, str) and residue.chain_code[0] in ("#", "@"):
            continue
        if isinstance(residue.sequence_code, str) and residue.sequence_code[0] in (
            "#",
            "@",
        ):
            continue
        result.append(residue)
    return result


def _parse_loops_residues(loop, chain_codes_to_select):

    offsets = [f"_{index}" for index in range(1, 16)]
    offsets.insert(0, "")
    index_sets = []
    for offset in offsets:
        if offset == "":
            linking_tag = f"linking{offset}"
            linking_index = (
                loop.tag_index(linking_tag) if linking_tag in loop.tags else None
            )

            residue_variant_tag = f"residue_variant{offset}"
            residue_variant_index = (
                loop.tag_index(residue_variant_tag)
                if residue_variant_tag in loop.tags
                else None
            )

            cis_peptide_tag = f"cis_peptide{offset}"
            cis_peptide_index = (
                loop.tag_index(cis_peptide_tag)
                if cis_peptide_tag in loop.tags
                else None
            )

        chain_code_tag = f"chain_code{offset}"
        chain_code_index = (
            loop.tag_index(chain_code_tag) if chain_code_tag in loop.tags else None
        )

        sequence_code_tag = f"sequence_code{offset}"
        sequence_code_index = (
            loop.tag_index(sequence_code_tag)
            if sequence_code_tag in loop.tags
            else None
        )

        residue_name_tag = f"residue_name{offset}"
        residue_name_index = (
            loop.tag_index(residue_name_tag) if residue_name_tag in loop.tags else None
        )

        if chain_code_index is None and sequence_code_index is None:
            continue

        index_set = {
            "chain_code_index": chain_code_index,
            "sequence_code_index": sequence_code_index,
            "residue_name_index": residue_name_index,
            "linking_index": linking_index,
            "residue_variant_index": residue_variant_index,
            "cis_peptide_index": cis_peptide_index,
        }

        index_sets.append(index_set)

    residues = []
    for line in loop:

        for index_set in index_sets:
            chain_code_index = index_set["chain_code_index"]
            sequence_code_index = index_set["sequence_code_index"]
            residue_name_index = index_set["residue_name_index"]
            linking_index = index_set["linking_index"]
            residue_variant_index = index_set["residue_variant_index"]
            cis_peptide_index = index_set["cis_peptide_index"]

            if chain_code_index is not None:
                chain_code = line[chain_code_index]
                if chain_code == UNUSED:
                    continue

                if not (
                    chain_codes_to_select is ANY_CHAIN
                    or chain_code in chain_codes_to_select
                ):
                    continue
            else:
                continue

            if sequence_code_index is not None:
                sequence_code = line[sequence_code_index]
                sequence_code = (
                    int(sequence_code) if is_int(sequence_code) else sequence_code
                )
            else:
                sequence_code = None

            residue_name = line[residue_name_index] if residue_name_index else None

            if linking_index is not None:
                linking = (
                    Linking[line[linking_index].upper()]
                    if line[linking_index] != NEF_UNKNOWN
                    else None
                )
            else:
                linking = None

            if cis_peptide_index is not None:
                cis_peptide = line[cis_peptide_index]
                if cis_peptide == UNUSED:
                    cis_peptide = False
                elif cis_peptide.lower() == "true":
                    cis_peptide = True
                elif cis_peptide.lower() == "false":
                    cis_peptide = False
                else:
                    cis_peptide = False
            else:
                cis_peptide = False

            if residue_variant_index is not None:
                residue_variants = line[residue_variant_index].split(",")
                residue_variants = (
                    ()
                    if residue_variants
                    == [
                        UNUSED,
                    ]
                    else tuple(residue_variants)
                )
            else:
                residue_variants = ()

            residue = SequenceResidue(
                chain_code=chain_code,
                sequence_code=sequence_code,
                residue_name=residue_name,
                linking=linking,
                is_cis=cis_peptide,
                variants=residue_variants,
            )
            if residue.chain_code and residue.sequence_code:
                residues.append(residue)

    return residues


def sequence_3let_to_res(
    sequence_3_let: List[str], chain_code: str, start: int = 1
) -> List[SequenceResidue]:
    """
    convert a list of 3 letter residue names to a list of sequence residues
    :param sequence_3_let: 3 letter names
    :param chain_code: chain code to use
    :param start:  start of the chain [default is 1]
    :return: a list of sequence residues
    """
    result = set()
    for res_number, residue_name in enumerate(sequence_3_let, start=start):
        result.add(SequenceResidue(chain_code, res_number, residue_name))

    return list(result)


def get_chain_starts(residues: List[SequenceResidue]) -> Dict[str, int]:
    """
    from a list of residues get the lowest residue number for each chain ignoring any sequence codes that
    can't be converted to an integer

    :param residues:  a list of residues from one or more chains
    :return: a dictionary of chain starts by chain_code
    """

    chain_residue_numbers = {}

    for residue in residues:

        if is_int(residue_number := residue.sequence_code):
            chain_residue_numbers.setdefault(residue.chain_code, []).append(
                residue_number
            )

    return {
        chain_code: min(residue_numbers)
        for chain_code, residue_numbers in chain_residue_numbers.items()
    }


def get_chain_ends(residues: List[SequenceResidue]) -> Dict[str, int]:
    """
    from a list of residues get the highest residue number for each chain ignoring any sequence codes that
    can't be converted to an integer

    :param residues:  a list of residues from one or more chains
    :return: a dictionary of chain starts by chain_code
    """

    chain_residue_numbers = {}

    for residue in residues:

        if is_int(residue_number := residue.sequence_code):
            chain_residue_numbers.setdefault(residue.chain_code, []).append(
                residue_number
            )

    return {
        chain_code: max(residue_numbers)
        for chain_code, residue_numbers in chain_residue_numbers.items()
    }


def get_chain_starts_and_ends(
    residues: List[SequenceResidue],
) -> Dict[str, Tuple[int, int]]:
    """
    from a list of residues get the lowest highest residue number for each chain ignoring any sequence codes that
    can't be converted to an integer

    :param residues:  a list of residues from one or more chains
    :return: a dictionary of chain starts and ends as a tuple by chain_code
    """

    chain_starts = get_chain_starts(residues)
    chain_ends = get_chain_ends(residues)

    result = {}
    for chain in chain_starts:
        result[chain] = (chain_starts[chain], chain_ends[chain])

    return result


def sequence_to_chains(residues: List[SequenceResidue]) -> List[str]:
    """
    from a list of residues get chain_codes

    :param residues:  a list of residues from one or more chains
    :return: a list of chain_codes
    """

    chain_codes = set()

    for residue in residues:

        chain_codes.add(residue.chain_code)

    return sorted(tuple(chain_codes))


def offset_chain_residues(
    residues: List[SequenceResidue], chain_residue_offsets: Dict[str, int]
) -> List[SequenceResidue]:
    """
    take a list of residues and offset residue sequence_codes by any offsets in chains if the
    sequence_code can be converted to an integer and the chain_code matches

    :param residues:  a list of residues from one or more chains
    :return: a list of residues
    """

    result = []
    for residue in residues:
        if residue.chain_code in chain_residue_offsets and is_int(
            residue.sequence_code
        ):
            offset = chain_residue_offsets[residue.chain_code]
            result.append(
                replace(residue, sequence_code=residue.sequence_code + offset)
            )
        else:
            result.append(residue)

    return result


def sequence_3let_list_from_sequence(
    sequence: List[SequenceResidue], count: int = 10, chain_code="A"
) -> List[str]:
    """
    convert a list of Residues to a set of lines containing space separated residue names
    [default 10] per line

    :param sequence: a sequence
    :param count: number of residues names to display per line
    :return: a list of lines as strings
    """

    residue_names = [
        residue.residue_name for residue in sequence if residue.chain_code == chain_code
    ]
    rows = chunks(residue_names, count)

    row_strings = []
    for row in rows:
        row_strings.append(" ".join(row))

    return row_strings


def make_chunked_sequence_1let(
    sequence_1_let: List[str], sub_chunk: int = 10, line_length: int = 100
) -> List[str]:
    """
    convert a list of strings into a chunked list of strings with (by default) every 100 sting being
    separated by a new line [line_length] and sequences of strings being separated by a space
    :param sequence_1_let: a list of strings typically 1 single letter
    :param sub_chunk: how often to separate strings by a space
    :param line_length: how many strings need to be seen before a new line is added
    :return: a list of formatted strings one per line
    """

    rows = chunks(sequence_1_let, line_length)

    row_strings = []
    for row in rows:
        row_chunks = list(chunks(row, sub_chunk))
        row_string = ["".join(chunk) for chunk in row_chunks]
        row_strings.append(" ".join(row_string))

    return row_strings


def replace_chain_in_atom_labels(
    atom_labels: List[AtomLabel], chain_code: str
) -> List[AtomLabel]:
    """
    replace the chain in a list of AtomLabels, not this operation is not inplace new AtomLabels are returned

    :param atom_labels: a list of AtomLabels
    :param chain_code: the new chain_code
    :return: the updated AtomLabels
    """
    result = []

    for selection in atom_labels:
        new_residue = replace(selection.residue, chain_code=chain_code)
        result.append(replace(selection, residue=new_residue))

    return result


def get_residue_name_from_lookup(
    chain_code: str,
    sequence_code: Union[str, int],
    lookup_table: Dict[Union[str, int], str],
) -> Optional[str]:
    """
    get a residue name from a chain_code and sequence_code, using the provided lookup table. The sequence code is
    treated being as either  a string or an int, strs are checked first
    :param chain_code: the chain code
    :param sequence_code: the residue code treated as both an int and a string
    :param lookup_table: the table of residues names indexed by chain_code_sequence_code
    :return:
    """

    seq_key = chain_code, str(sequence_code)

    residue_name = lookup_table.get(seq_key, None)

    if not residue_name:
        if is_int(sequence_code):
            seq_key = chain_code, int(sequence_code)

    residue_name = lookup_table.get(seq_key, UNUSED)

    if residue_name is not None:
        residue_name = residue_name.upper()

    return residue_name


# TODO sequence codes should be ints if they look like ints (currently fails tests)
def sequence_to_residue_name_lookup(
    sequence: List[SequenceResidue],
) -> Dict[Tuple[str, int], str]:
    """
    build a lookup table from sequence_code, chain_code to residue_type from a list of sequence residues
    :param sequence: a list of sequence residues
    :return: the lookup table
    """

    result: Dict[Tuple[str, int], str] = {}
    if len(sequence) > 0:
        result = {
            (residue.chain_code, str(residue.sequence_code)): residue.residue_name
            for residue in sequence
        }

    return result


def atom_sort_key(item: AtomLabel) -> Tuple[Any, ...]:
    """

    key based sorter for atom labels that enables them to be sorted in a sensible order including sequence codes
    which are not numeric...

    :param item: an atom label
    :return: a tuple based on the atom label which gives the correct sort order
    """

    try:
        items = []
        residue = item.residue
        chain_code = residue.chain_code.strip()
        prefix, numbers = strip_characters_right(chain_code, string.digits)

        if numbers == "":
            numbers = 0, ""
        if is_int(numbers):
            numbers = int(numbers), ""
        else:
            numbers = 0, numbers

        if UNUSED in prefix:
            order = 2
        elif "@" in prefix or "#" in prefix:
            order = 1
        else:
            order = 0

        items.append([order, prefix, *numbers])

        sequence_code = str(residue.sequence_code).strip()
        prefix, numbers = strip_characters_right(sequence_code, (string.digits + "+-"))

        if is_int(numbers):
            numbers = [int(numbers), "", 0]
        elif "+" in numbers or "-" in numbers:
            for splitter in "+-":
                if splitter in numbers:
                    numbers = numbers.rsplit(splitter, maxsplit=1)
                    numbers = [numbers[0], splitter, numbers[1]]
                    break
            if is_int(numbers[0]) and is_int(numbers[-1]):
                numbers[0] = int(numbers[0])
                numbers[-1] = int(numbers[-1])
        else:
            numbers = [0, numbers, 0]

        if UNUSED in prefix:
            order = 2
        elif "@" in prefix or "#" in prefix:
            order = 1
        else:
            order = 0

        items.append([order, prefix, *numbers])

        items.append(residue.residue_name.lower())
        items.append(item.atom_name.lower())
    except Exception as e:
        msg = f"""
            I can't sort {item} because it doesn't seem to be behaving like an AtomLabel...
            [{e}]
        """
        raise ValueError(msg)

    return items


def exit_if_chain_not_in_entrys_sequence(chain_code: str, entry: Entry):

    nef_sequence = sequence_from_entry(entry)

    _exit_if_chain_not_in_sequence(chain_code, nef_sequence, entry)


def _exit_if_chain_not_in_sequence(
    chain_code: str,
    nef_sequence: List[SequenceResidue],
    entry: Entry,
):
    file_name = f"{entry.source} [stdin]" if entry.source == "-" else entry.source

    nef_chain_codes = sequence_to_chains(nef_sequence)
    if chain_code and chain_code not in nef_chain_codes:
        msg = f"""
            The chain code {chain_code} was not found in the input chain codes in the nef stream
            with entry name {entry.entry_id}  in file {file_name}
            the chain codes found were {','.join(nef_chain_codes)}
        """

        exit_error(msg)



from collections import OrderedDict
from dataclasses import replace
from pathlib import Path
from typing import Dict, List, NoReturn, Tuple, Union

import tabulate
import typer
from pynmrstar import Saveframe

from nef_pipelines.lib.nef_lib import (
    UNUSED,
    loop_row_dict_iter,
    read_entry_from_file_or_stdin_or_exit_error,
    select_frames_by_name,
)
from nef_pipelines.lib.sequence_lib import sequence_from_entry, sequence_to_chains
from nef_pipelines.lib.structures import AtomLabel, RdcRestraint, SequenceResidue
from nef_pipelines.lib.translation_lib import IUPAC_XPLOR, translate_atom_label
from nef_pipelines.lib.util import (
    STDIN,
    exit_error,
    flatten,
    get_display_file_name,
    read_float_or_exit,
)
from nef_pipelines.transcoders.pales import export_app

app = typer.Typer()

# TODO: name translations
# TODO: correct weights
# TODO: move utilities to lib
# TODO: support multiple chains

NEF_RDC_RESTRAINT_FRAME_CATEGORY = "nef_rdc_restraint_list"
NEF_RDC_RESTRAINT_CATEGORY = "nef_rdc_restraint"


# noinspection PyUnusedLocal
@export_app.command()
def rdcs(
    user_chains: List[str] = typer.Option(
        [],
        "-c",
        "--chain",
        help="chains to export, to add mutiple chains use repeated calls  [default: all the chains in the rdc restraint"
        "frames]",
        metavar="<CHAIN-CODE>",
    ),
    file_name_template: str = typer.Option(
        "%s.out",
        "-t",
        "--template",
        help="the template for the filename to export to %s will get replaced by the name of the chain or a filename if"
        "set with the sequence_text-names option",
        metavar="<sequence-sequence_text.seq>",
    ),
    output_to_stdout: bool = typer.Option(
        False, "-o", "--out", help="write the files to stdout for debugging"
    ),
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--input",
        help="input to read NEF data from [- is stdin]",
    ),
    raw_weights: List[str] = typer.Option(
        [],
        "-w",
        "--weight",
        help="weights for rdcs as a comma separated triple of 2 atom names and a weight [no spaces e.g HN,N,1.0],"
        " multiple weights may be added by repeated calls, efault is HN,N,1.0",
    ),
    segids: bool = typer.Option(
        False,
        "-s",
        "--segids",
        help="include segids in the file",
    ),
    default_uncertainty: float = typer.Option(
        1.0,
        "-u",
        "--uncertainty",
        help="the uncertainty to use if none is provided",
    ),
    frame_selectors: List[str] = typer.Option(
        [],
        "-f",
        "--frame",
        help="selector for the rdc restraint frame to use, can be called multiple times and include wild cards",
    ),
):
    """- convert nef rdc restraints to pales"""

    entry = read_entry_from_file_or_stdin_or_exit_error(input)

    weights = _parse_weights(raw_weights)

    sequence = sequence_from_entry(entry)

    sequence_chains = sequence_to_chains(sequence)

    rdc_frames = entry.get_saveframes_by_category(NEF_RDC_RESTRAINT_FRAME_CATEGORY)

    rdc_restaints_by_frame = {
        frame.name: _rdc_restraints_from_frame(frame, weights) for frame in rdc_frames
    }

    rdc_chains = [
        _rdc_restraints_to_chains(rdc_frame)
        for rdc_frame in rdc_restaints_by_frame.values()
    ]

    rdc_chains = flatten(rdc_chains)

    if not frame_selectors:
        frame_selectors = ["*"]

    rdc_frames = select_frames_by_name(rdc_frames, frame_selectors)

    if not user_chains:
        user_chains = set(rdc_chains).intersection(set(sequence_chains))

    _exit_if_user_chains_not_in_restraints(rdc_chains, user_chains, input, rdc_frames)

    for frame_name, restraints in rdc_restaints_by_frame.items():

        _exit_if_mixed_chain_restraints(restraints, frame_name, input)

        restraints = _translate_atom_names(restraints, IUPAC_XPLOR)

        restraints = _update_restraint_uncertainty(restraints, default_uncertainty)

        print()

        _print_restraints(restraints, weights, segids)


def _translate_atom_names(
    restraints: List[RdcRestraint], naming_scheme=IUPAC_XPLOR
) -> List[RdcRestraint]:
    result = []
    for restraint in restraints:
        atom_1 = restraint.atom_1
        atom_2 = restraint.atom_2

        atom_1 = translate_atom_label(atom_1, naming_scheme)
        atom_2 = translate_atom_label(atom_2, naming_scheme)

        restraint = replace(restraint, atom_1=atom_1)
        restraint = replace(restraint, atom_2=atom_2)
        restraint.atom_1 = atom_1

        result.append(restraint)

    return result


def _rdc_restraints_from_frame(frame: Saveframe, weights: Dict[Tuple[str, str], float]):
    result = []

    for row in loop_row_dict_iter(frame.loops[0]):
        atom_1 = AtomLabel(
            SequenceResidue(
                row["chain_code_1"],
                int(row["sequence_code_1"]),
                row["residue_name_1"],
            ),
            row["atom_name_1"],
        )
        atom_2 = AtomLabel(
            SequenceResidue(
                row["chain_code_2"],
                int(row["sequence_code_2"]),
                row["residue_name_2"],
            ),
            row["atom_name_2"],
        )

        weight_key = _build_weights_key(row["atom_name_1"], row["atom_name_2"])
        # this should be
        weight = weights[weight_key] if weight_key in weights else 1.0
        rdc = RdcRestraint(
            atom_1,
            atom_2,
            row["target_value"],
            row["target_value_uncertainty"],
            weight,
        )

        result.append(rdc)

    return result


def _parse_weights(raw_weights):

    result = {}
    for raw_weight in raw_weights:

        raw_weight_fields = raw_weight.split(",")

        if len(raw_weight_fields) != 3:
            exit_error(
                f"bad weight {raw_weight} weights should have 3 fields separated by commas [no spaces]"
                f" atom_1,atom_2,weight. e.g HN,N,1.0"
            )

        atom_1, atom_2, raw_rdc_weight = raw_weight_fields

        rdc_weight = read_float_or_exit(
            raw_rdc_weight,
            message=f"weights should have 3 fields separated by commas [no spaces] atom_1,atom_2,weight weight should"
            f" be a float i got {raw_weight} for field feild [{raw_rdc_weight}] which is not a float",
        )

        key = _build_weights_key(atom_1, atom_2)
        result[key] = rdc_weight

    # it not possibe to add default weights?
    HN_N_key = _build_weights_key("HN", "N")
    if HN_N_key not in result:
        result[HN_N_key] = 1.0

    return result


def _build_weights_key(atom_1, atom_2):
    return tuple(sorted([atom_1, atom_2]))


def _print_restraints(
    restraints: List[RdcRestraint],
    weights: Dict[Tuple[str, str], float],
    use_segids: bool = True,
):

    segid_i = "SEGNAME_I" if use_segids else ""
    segid_i_format = "%4s" if use_segids else ""
    segid_j = "SEGNAME_J" if use_segids else ""
    segid_j_format = "%4s" if use_segids else ""

    VARS_1 = f"VARS   {segid_i}        RESID_I RESNAME_I ATOMNAME_I {segid_j}         RESID_J RESNAME_J ATOMNAME_J"
    FORMAT_1 = f"FORMAT {segid_i_format} %5d     %6s       %6s        {segid_j_format}  %5d     %6s       %6s       "

    VARS_2 = "D     DD     W   "
    FORMAT_2 = "%9.3f %9.3f  %.2f"

    VARS = " ".join([VARS_1, VARS_2]).strip().split()
    FORMAT = " ".join([FORMAT_1, FORMAT_2]).strip().split()

    table = [VARS, FORMAT]

    for i, restraint in enumerate(restraints):
        atom_1 = restraint.atom_1
        atom_2 = restraint.atom_2

        weights_key = _build_weights_key(atom_1.atom_name, atom_2.atom_name)
        weight = weights[weights_key]
        row_data = OrderedDict(
            [
                ("chain_code_1", atom_1.residue.chain_code),
                ("sequence_code_1", atom_1.residue.sequence_code),
                ("residue_name_1", atom_1.residue.residue_name),
                ("atom_name_1", atom_1.atom_name),
                ("chain_code_2", atom_2.residue.chain_code),
                ("sequence_code_2", atom_2.residue.sequence_code),
                ("residue.residue_name", atom_2.residue.residue_name),
                ("atom_name", atom_2.atom_name),
                ("target_value", restraint.value),
                ("rdc_error", restraint.value_uncertainty),
                ("weight", weight),
            ]
        )
        row = [""]
        for key, value in row_data.items():
            if "chain_code" not in key:
                row.append(value)
            elif "chain_code" in key and use_segids:
                row.append(value)

        table.append(row)
    print()
    print(tabulate.tabulate(table, tablefmt="plain"))


def _rdc_restraints_to_chains(restrants: List[RdcRestraint]) -> List[str]:
    chains = set()

    for restraint in restrants:

        chains.add(restraint.atom_1.residue.chain_code)
        chains.add(restraint.atom_2.residue.chain_code)

    if UNUSED in chains:
        chains.remove(UNUSED)

    return sorted(list(chains))


def _exit_if_user_chains_not_in_restraints(
    rdc_chains: List[str],
    user_chains: List[str],
    input: Path,
    rdc_frames: List[Saveframe],
) -> Union[None, NoReturn]:
    rdc_chains = set(rdc_chains)
    user_chains = set(user_chains)
    if not rdc_chains.issubset(user_chains):
        missing_chains = user_chains - rdc_chains
        rdc_frame_names = ", ".join([frame.name for frame in rdc_frames])
        msg = f"""\
            some of the selected chains are not present in the restraints in the input nef stream
            the chains selected were: {', '.join(user_chains)}
            the chains in the input were: {', '.join(rdc_chains)}
            the missing chains were: {', '.join(missing_chains) }
            input was: {get_display_file_name(input)}
            rdc_frames_were: {rdc_frame_names}
        """
        exit_error(msg)


def _exit_if_mixed_chain_restraints(
    restraints: List[RdcRestraint], frame_name, input
) -> Union[NoReturn, None]:

    for row, restraint in enumerate(restraints, start=1):

        chain_1 = restraint.atom_1.residue.chain_code
        chain_2 = restraint.atom_2.residue.chain_code
        if chain_1 != chain_2:
            msg = f"""\
                in row {row} of frame {frame_name} in {get_display_file_name(input)}
                the two chains have different chain ids [not currently supported]
                chain 1: {chain_1}, chain_2: {chain_2}
            """

            exit_error(msg)


def _update_restraint_uncertainty(
    restraints: List[RdcRestraint], uncertainty: float
) -> List[RdcRestraint]:
    result = []

    for restraint in restraints:
        if restraint.value_uncertainty == UNUSED:
            result.append(replace(restraint, value_uncertainty=uncertainty))
        else:
            result.append(restraint)

    return result



from dataclasses import replace
from itertools import islice
from pathlib import Path
from typing import List

import typer
from typer import Argument, Option

from nef_pipelines.lib.nef_lib import read_entry_from_file_or_stdin_or_exit_error
from nef_pipelines.lib.sequence_lib import (
    chains_from_frames,
    get_chain_code_iter,
    sequence_to_nef_frame,
    sequences_from_frames,
)
from nef_pipelines.lib.util import exit_error, parse_comma_separated_options
from nef_pipelines.tools.chains import chains_app

app = typer.Typer()


# noinspection PyUnusedLocal
@chains_app.command()
def clone(
    input_path: Path = typer.Option(
        None,
        metavar="|PIPE|",
        help="file to read NEF data from default is stdin '-'",
    ),
    target: str = Argument("A", help="chain to clone"),
    count: int = Argument(1, help="how many copys to make"),
    chain_codes: List[str] = Option(
        None,
        "-c",
        "--chains",
        help="new chain codes to add otherwise defaults to next available ascii upper case letter,"
        " can be called multiple times or be a comma separated list",
    ),
):
    """- duplicate chains one or more times"""

    if not count > 0:
        exit_error(f"clone count must be > 0 i got: {count}")

    chain_codes = parse_comma_separated_options(chain_codes)

    entry = read_entry_from_file_or_stdin_or_exit_error(input_path)

    molecular_system = entry.get_saveframes_by_category("nef_molecular_system")

    if len(molecular_system) < 1:
        exit_error("couldn't find a molecular system frame in the stream")

    if len(molecular_system) > 1:
        exit_error(
            "found more than one molecular system frame, this is not a valid nef file!"
        )

    molecular_system = molecular_system[0]

    sequence = sequences_from_frames(molecular_system)

    # this should come from a sequence not a frame...
    existing_chain_codes = chains_from_frames(molecular_system)

    if target not in existing_chain_codes:
        exit_error(
            f"couldn't find target chain {target} in {', '.join(existing_chain_codes)}"
        )
    useable_chain_codes = get_chain_code_iter(
        chain_codes, [target, *existing_chain_codes]
    )
    new_chain_codes = islice(useable_chain_codes, count)

    target_residues = [residue for residue in sequence if residue.chain_code == target]

    for new_chain_code in new_chain_codes:
        for residue in target_residues:
            new_residue = replace(residue, chain_code=new_chain_code)
            sequence.append(new_residue)

    sequence = sorted(sequence)

    molecular_system_frame = sequence_to_nef_frame(sequence)

    entry.remove_saveframe("nef_molecular_system")

    entry.add_saveframe(molecular_system_frame)

    print(entry)



from dataclasses import replace
from pathlib import Path
from typing import Iterable

import typer

from nef_pipelines.lib.nef_lib import (
    add_frames_to_entry,
    read_file_or_exit,
    read_or_create_entry_exit_error_on_bad_file,
)
from nef_pipelines.lib.sequence_lib import translate_1_to_3
from nef_pipelines.lib.shift_lib import shifts_to_nef_frame
from nef_pipelines.lib.structures import ShiftList
from nef_pipelines.lib.util import STDIN
from nef_pipelines.transcoders.nmrpipe import import_app
from nef_pipelines.transcoders.nmrpipe.nmrpipe_lib import (
    read_db_file_records,
    read_shift_file,
)

app = typer.Typer()


# noinspection PyUnusedLocal
@import_app.command()
def shifts(
    chain_code: str = typer.Option(
        "A",
        "--chains",
        help="chain codes, can be called multiple times and or be a comma separated list [no spaces!]",
        metavar="<CHAIN-CODES>",
    ),
    entry_name: str = typer.Option("nmrpipe", help="a name for the entry"),
    frame_name: str = typer.Option("nmrpipe", help="a name for the frame"),
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--input",
        metavar="|PIPE|",
        help="pipe to read NEF data from, for testing [overrides stdin !use stdin instead!]",
    ),
    file_name: Path = typer.Argument(
        ..., help="input files of type nmrpipe.tab", metavar="<TAB-FILE>"
    ),
):
    """convert nmrpipe shift file <nmrpipe>.tab files to NEF"""

    entry = read_or_create_entry_exit_error_on_bad_file(input, entry_name)

    lines = read_file_or_exit(file_name)

    entry = pipe(entry, lines, chain_code, frame_name, file_name)

    print(entry)


def pipe(entry, lines, chain_code, frame_name, file_name):
    nmrpipe_frames = []

    nmrpipe_shifts = read_shifts(lines, chain_code, file_name)

    nmrpipe_shifts = _convert_1let_3let_if_needed(nmrpipe_shifts)

    frame = shifts_to_nef_frame(nmrpipe_shifts, frame_name)

    nmrpipe_frames.append(frame)

    entry = add_frames_to_entry(entry, nmrpipe_frames)

    return entry


def _convert_1let_3let_if_needed(shifts: ShiftList) -> ShiftList:

    new_shifts = []
    for shift in shifts.shifts:
        residue_name = [
            shift.atom.residue.residue_name,
        ]

        if len(residue_name[0]) == 1:
            residue_name = translate_1_to_3(residue_name)

        new_residue = replace(shift.atom.residue, residue_name=residue_name[0])
        new_atom = replace(shift.atom, residue=new_residue)
        new_shift = replace(shift, atom=new_atom)

        new_shifts.append(new_shift)

    new_shift_list = ShiftList(new_shifts)

    return new_shift_list


def read_shifts(
    shift_lines: Iterable[str],
    chain_code: str = "A",
    file_name: str = "unknown",
) -> ShiftList:

    gdb_file = read_db_file_records(shift_lines, file_name)

    return read_shift_file(gdb_file, chain_code)



from math import sqrt
from pathlib import Path
from typing import List

import typer
from numpy import mean
from ordered_set import OrderedSet
from pynmrstar import Entry, Saveframe

from nef_pipelines.lib.nef_frames_lib import NEF_PIPELINES_NAMESPACE
from nef_pipelines.lib.nef_lib import read_entry_from_file_or_stdin_or_exit_error
from nef_pipelines.lib.shift_lib import IntensityMeasurementType
from nef_pipelines.lib.util import exit_error, parse_comma_separated_options
from nef_pipelines.tools.fit import fit_app
from nef_pipelines.tools.fit.fit_lib import (
    NEFPLSFitLibException,
    _combine_relaxation_series,
    _exit_if_no_frame_selectors,
    _exit_if_no_series_frames_selected,
    _fit_results_as_frame,
    _select_relaxation_series_or_exit,
    _series_frame_to_id_series_data,
    _series_frame_to_outputs,
    calculate_noise_level_from_replicates,
)

try:
    from streamfitter.interface import LoggingLevels, NoiseInfo, NoiseInfoSource

except ImportError as e:

    stream_fitter_import_error = str(e)

VERBOSE_HELP = """
how verbose to be, each call of verbose increases the verbosity, note this currently only reports JAX warnings
"""


def _chunker(seq, size):
    return (seq[pos : pos + size] for pos in range(0, len(seq), size))


@fit_app.command()
def t1noe(
    input: Path = typer.Option(
        None,
        "-i",
        "--in",
        metavar="NEF-FILE",
        help="read NEF data from a file instead of stdin",
    ),
    cycles: int = typer.Option(
        1000,
        "-c",
        "--cycles",
        help="number of cycles for error propogation",
    ),
    noise_level: float = typer.Option(
        None,
        "-n",
        "--noise-level",
        help="noise level to use instead of value from replicates",
    ),
    seed: int = typer.Option(
        42, "-s", "--seed", help="seed for random number generator"
    ),
    data_type: IntensityMeasurementType = typer.Option(
        IntensityMeasurementType.HEIGHT, "-d", "--data-type", help="data type to fit"
    ),
    verbose: int = typer.Option(LoggingLevels.WARNING, count=True, help=VERBOSE_HELP),
    frames_selectors: List[str] = typer.Argument(
        None, help="select frames to fit, these must come in pairs"
    ),
    outputs: List[str] = typer.Option(
        None,
        "--outputs",
        help="""\
            a list of output relaxation list names, there should be two in the order R1 and then NOE, these
            override values in the input frames
        """,
    ),
):
    """- fit pairs of series to exponential decays with a shared R1, amplitude and asymtote  with error propagation
    to measure T1s and {1H}–15N nOes as described in 'TROSY pulse sequence for simultaneous measurement of the
    15N R1 and {1H}–15N NOE in deuterated proteins' O’Brien & Palmer doi://10.1007/s10858-018-0181-6 [alpha]
    """

    entry = read_entry_from_file_or_stdin_or_exit_error(input)

    frame_selectors = parse_comma_separated_options(frames_selectors)

    if not frames_selectors:
        series_frames = entry.get_saveframes_by_category("nef_series_list")
        if len(series_frames) == 1:
            frame_selectors = [
                series_frames[0].name,
            ]

    _exit_if_no_frame_selectors(frame_selectors)

    series_frames = _select_relaxation_series_or_exit(entry, frames_selectors)

    _exit_if_series_frame_not_in_pairs(series_frames)

    _exit_if_no_series_frames_selected(series_frames, frame_selectors)

    if outputs:
        outputs = parse_comma_separated_options(outputs)

    # TODO outputs should be a list of pairs, isolate in function
    if outputs and len(outputs) != 2:
        joined_outputs = ", ".join(outputs)
        msg = f"""
            when fitting symmetrical R1 curves and {{1H}}-15N nOes there must be two output relaxation lists
            the first will contain R1 fits and the second {{1H}}-15N nOes, you gave {len(outputs)} outputs
            which were:

            {joined_outputs}
        """

        exit_error(msg)

    entry = pipe(
        entry, series_frames, cycles, noise_level, data_type, seed, verbose, outputs
    )

    print(entry)


def pipe(
    entry: Entry,
    series_frames: List[Saveframe],
    cycles: int,
    noise_level,
    data_type: IntensityMeasurementType,
    seed: int,
    verbose: int = 0,
    outputs=None,
) -> Entry:

    try:
        from streamfitter import fitter

        if fitter:

            function = fitter.get_function(
                fitter.FUNCTION_TWO_EXPONENTIAL_DECAYS_2_PAMETER_SHARED_RATE
            )
        else:
            raise ImportError(stream_fitter_import_error)
    except ImportError as e:

        msg = f"""
                error the package streamfitter is not installed or is not importing properly
                the error was {e}
            """

        exit_error(msg)

    for series_frame_1, series_frame_2 in _chunker(series_frames, 2):

        id_series_data_1 = _series_frame_to_id_series_data(
            series_frame_1, NEF_PIPELINES_NAMESPACE, entry
        )

        id_series_data_2 = _series_frame_to_id_series_data(
            series_frame_2, NEF_PIPELINES_NAMESPACE, entry
        )

        id_outputs_1 = _series_frame_to_outputs(
            series_frame_1, NEF_PIPELINES_NAMESPACE, entry
        )

        id_outputs_2 = _series_frame_to_outputs(
            series_frame_2, NEF_PIPELINES_NAMESPACE, entry
        )

        if noise_level is not None:
            requested_noise_source = NoiseInfoSource.CLI
            noise_source = NoiseInfoSource.CLI
            noise_error_fraction = None
            num_replicates = None
        else:
            requested_noise_source = NoiseInfoSource.REPLICATES
            noise_level_1, noise_1_error_fraction, num_replicates_1 = (
                calculate_noise_level_from_replicates(id_series_data_1)
            )

            noise_level_2, _noise_2_error_fraction, num_replicates_2 = (
                calculate_noise_level_from_replicates(id_series_data_2)
            )

            if noise_level_1 and noise_level_2:
                noise_level = sqrt(noise_level_1**2 + noise_level_2**2)
                noise_error_fraction = (
                    mean(noise_1_error_fraction, noise_1_error_fraction) * 1 / sqrt(2)
                )
                noise_source = NoiseInfoSource.REPLICATES
            else:
                noise_level = None
                noise_error_fraction = None
                noise_source = NoiseInfoSource.NONE
            num_replicates = num_replicates_1 + num_replicates_2

        if num_replicates == 0:
            noise_source = NoiseInfoSource.NONE

        noise_info = NoiseInfo(
            noise_source,
            noise_level,
            noise_error_fraction,
            num_replicates,
            requested_noise_source,
        )

        print("#", noise_info)

        data_ids = OrderedSet([*id_series_data_1.keys(), *id_series_data_2.keys()])

        if not outputs:
            outputs = OrderedSet()
            for data_id in data_ids:
                outputs.update(id_outputs_1[data_id])
                outputs.update(id_outputs_2[data_id])
            outputs = list(outputs)

        outputs = [output.replace("nefpls_relaxation_list_", "") for output in outputs]

        try:

            id_series_data = {
                data_id: _combine_relaxation_series(
                    id_series_data_1[data_id], id_series_data_2[data_id]
                )
                for data_id in data_ids
            }
        except NEFPLSFitLibException as e:
            msg = f"""
                merging the series {series_frame_1.name} and {series_frame_2.name} in entry {entry.entry_id}
                failed because {e}
            """
            exit_error(msg)

        id_xy_data = {
            data_id: (series_datum.variable_values, series_datum.values)
            for data_id, series_datum in id_series_data.items()
        }

        results = fitter.fit(
            function(),
            id_xy_data,
            cycles,
            noise_info,
            seed,
            verbose=verbose,
        )

        fits = results["fits"]
        monte_carlo_errors = results["monte_carlo_errors"]
        monte_carlo_value_stats = results["monte_carlo_value_stats"]
        monte_carlo_param_values = results["monte_carlo_param_values"]
        noise_level = results["noise_level"]
        version_strings = results["versions"]

        fit_ids = ["time_constant", "offset"]
        for fit_name, output in zip(fit_ids, outputs):
            frame = _fit_results_as_frame(
                series_frame_1,
                NEF_PIPELINES_NAMESPACE,
                entry,
                fits,
                fit_name,
                output,
                monte_carlo_errors,
                monte_carlo_value_stats,
                monte_carlo_param_values,
                seed,
                noise_info,
                version_strings,
                "r1-noe-symmetric-double-exponential",
            )

            entry.add_saveframe(frame)

    return entry


def _exit_if_series_frame_not_in_pairs(series_frames):

    if len(series_frames) % 2 != 0:
        frame_names = [
            "{i}. {frame.name}" for i, frame in enumerate(series_frames, start=1)
        ]
        msg = f"""
            This fitter requires pairs of data to fit pairs of exponentials you provided and odd number of data sets:

            {frame_names}
        """
        exit_error(msg)


# implementation thoughts
# 1. having a spectrometer frequency and field strength in the same construct is a pain, what do we define as the
#    conversion ratio i took value in cavanagh and cross-checked it with from https://www.kherb.io/docs/nmr_table.html
#    which uses the codata value for 1H [but not for 15N and 13C which use the IAEA values]
# 2. if we have 1H and 15N and 13C spectrometer frequencies which should we use to
#    calculate the field strength? i chose the one with the highest gamma, but note my gamma ratios do not match
#    those in https://www.kherb.io/docs/nmr_table.html exacly
# 3. do we need the value type should it be Nz for a 15N experiment, what are I and S defined as in this context
# 4. how do we determine the relaxation_atom_id shouldn't also be available on the series data or come from the series
#    data as it's the type of experiment and the experimenta series that determine this, currently choose lowest gamma!
# 5. the value doesn't have a type just a unit, should it?
# 6. the source maybe should be on the data series rather than the relaxation list -  it knoes what its is
#    the relaxation list doesn't need tow
# 7. the nef_relaxation doesn't leave room for other fit values e.g I or the baseline offset for a 3 point fit...
# 8. room for a comment on each fit would be good...
# 9. when would the index and data_id in a nef_relaxation_list be different?
# 10. we should say the value error is sigma
# 11. why are some of our constants capitalised and others not - we should be consistent
#     e.g. experimental | simulated | theoretical vs OnePhaseDecay | ExponentialDecay | InversionRecovery | other
# 12. why is relaxation_list_id repeated when nef_series_list isn't a singleton and applies to only one experimental
# series?
# 13. question can you use _nef_series_data as an input table for a calculation and leave relaxation_list_id blank to be
# filled out by the calculation engine. Also for het noe should the series list values be  bool or 1 and 0 for the
# saturation
# 14. not sure not repeating the atom names in the series data is a good idea as the _nef_relaxation loop needs
# them for the result and you have to wade through all the spectra again while dou don't need to for the actual dat and
# they can be inconsistent or even missing...
# 15. there is no connection from the nef_relaxation_list to the original nef_series_list specifically
#     it has a data_id but series list that data_id refers to isn't mentioned... to find it i have to scan
#     through all save_nef_series_lists and what happens if 2 refer to this nef_relaxation_list...
# 16 experiment_ypes homonuclear_NOEs is plural and some are singular
# 17. shouldn't it be value_uncertainty



import copy

_globals = {}


def set_global(key, value):
    _globals[key] = value


def get_global(key, default):
    return _globals[key] if key in _globals else default


def debug_clear_globals():
    _globals.clear()


def debug_get_globals():
    return copy.deepcopy(_globals)



import pytest
from frozendict import frozendict

from nef_pipelines.lib.test_lib import path_in_test_data, read_test_data
from nef_pipelines.transcoders.rcsb.rcsb_lib import (
    PdbSecondaryStructureType,
    RCSBFileType,
    SequenceSource,
    guess_cif_or_pdb,
    parse_cif,
    parse_pdb,
)

FILE_TYPE_TO_READER = {"cif": parse_cif, "pdb": parse_pdb}


@pytest.mark.parametrize(
    "file_name",
    [
        # ("1l2y_short.pdb"),
        ("1l2y_short.cif"),
    ],
)
def test_short(file_name):

    file_path = path_in_test_data(__file__, file_name)
    file_type = file_name.split(".")[-1]

    reader = FILE_TYPE_TO_READER[file_type]

    with open(file_path) as lines:
        structure = reader(lines, file_name)

    assert len(structure.sequences) == 1
    assert len(structure.secondary_structure) == 1
    assert len(structure.models) == 2

    model_0 = structure.models[0]
    model_1 = structure.models[1]
    assert len(model_0.chains) == len(model_1.chains) == 1

    model_0_chain_A = model_0.chains["A"]
    model_1_chain_A = model_1.chains["A"]
    assert len(model_0_chain_A.residues) == len(model_1_chain_A.residues) == 4

    assert model_0_chain_A.residues[0] != model_1_chain_A.residues[0]

    assert list(structure.sequences.keys()) == [
        1,
    ]
    assert structure.sequences[1].id == 1
    assert structure.sequences[1].residues == ["ASN", "LEU", "TYR", "ILE"]
    assert structure.sequences[1].structure is structure

    secondary_structure = structure.secondary_structure["A"][0]
    assert secondary_structure.chain_code == "A"
    assert secondary_structure.start_sequence_code == 1
    assert secondary_structure.end_sequence_code == 4
    assert secondary_structure.alternative_location is None
    assert (
        secondary_structure.secondary_structure_type
        == PdbSecondaryStructureType.HELIX_RIGHT_HANDED_ALPHA
    )
    assert secondary_structure.structure == structure

    EXPECTED_ATOMS = {
        frozendict(
            {
                "model": 1,
                "chain_code": "A",
                "sequence_code": 1,
                "residue_name": "ASN",
                "atom_number": 0,
            }
        ): {
            "atom_name": "N",
            "x": -8.901,
            "y": 4.127,
            "z": -0.555,
            "element": "N",
        },
        frozendict(
            {
                "model": 1,
                "chain_code": "A",
                "sequence_code": 3,
                "residue_name": "TYR",
                "atom_number": 19,
            }
        ): {
            "atom_name": "HE2",
            "x": 0.033,
            "y": 4.952,
            "z": 4.233,
            "element": "H",
        },
        frozendict(
            {
                "model": 2,
                "chain_code": "A",
                "sequence_code": 2,
                "residue_name": "LEU",
                "atom_number": 4,
            }
        ): {
            "atom_name": "CB",
            "x": -4.073,
            "y": 4.863,
            "z": -3.128,
            "element": "C",
        },
        frozendict(
            {
                "model": 2,
                "chain_code": "A",
                "sequence_code": 4,
                "residue_name": "ILE",
                "atom_number": 18,
            }
        ): {
            "atom_name": "HD13",
            "x": -9.973,
            "y": 1.875,
            "z": 1.360,
            "element": "H",
        },
    }

    for key_values, data_values in EXPECTED_ATOMS.items():
        model = structure.models[key_values["model"] - 1]
        chain = model.chains[key_values["chain_code"]]
        residue = chain.residues[key_values["sequence_code"] - 1]
        atom = residue.atoms[key_values["atom_number"]]

        assert model.serial == key_values["model"]
        assert chain.chain_code == key_values["chain_code"]
        assert residue.chain.chain_code == key_values["chain_code"]
        assert atom.residue.residue_name == key_values["residue_name"]
        for key, value in data_values.items():
            assert getattr(atom, key) == value


def test_file_type_from_extension():
    assert guess_cif_or_pdb("", "test.mmcif") == RCSBFileType.CIF
    assert guess_cif_or_pdb("", "test.pdb") == RCSBFileType.PDB


@pytest.mark.parametrize(
    "file_name",
    [
        ("1l2y_short.pdb"),
        ("1l2y_short.cif"),
    ],
)
def test_file_type_from_data(file_name):

    file_type_str = file_name.split(".")[-1]

    file_data = read_test_data(file_name, __file__).split("\n")
    assert guess_cif_or_pdb(file_data) == RCSBFileType[file_type_str.upper()]


@pytest.mark.parametrize(
    "file_name",
    [
        ("1k0o.pdb"),
        ("1k0o.cif"),
    ],
)
def test_parse_clic1(file_name):
    file_data = read_test_data(file_name, __file__).split("\n")
    if file_name.endswith(".pdb"):
        structure = parse_pdb(file_data, file_name)
    elif file_name.endswith(".cif"):
        structure = parse_cif(file_data, file_name)
    else:
        raise Exception(f"unexpected file type {file_name.split('.')[-1]}")

    assert len(structure.sequences) == 1
    assert len(structure.secondary_structure) == 2
    assert list(structure.secondary_structure.keys()) == ["A", "B"]
    assert len(structure.models) == 1

    model_0 = structure.models[0]
    model_0_chain_A = model_0.chains["A"]
    model_0_chain_B = model_0.chains["B"]

    assert len(model_0_chain_A.residues) == 225
    assert len(model_0_chain_B.residues) == 213

    assert list(structure.sequences.keys()) == [
        1,
    ]
    assert structure.sequences[1].id == 1
    assert len(structure.sequences[1].residues) == 241
    assert structure.sequences[1].structure is structure

    for chain in "AB":
        secondary_structure = structure.secondary_structure[chain][0]
        assert (
            secondary_structure.secondary_structure_type
            == PdbSecondaryStructureType.SHEET
        )
        assert secondary_structure.start_sequence_code == 8
        assert secondary_structure.end_sequence_code == 13

        secondary_structure = structure.secondary_structure[chain][-1]
        assert (
            secondary_structure.secondary_structure_type
            == PdbSecondaryStructureType.HELIX_RIGHT_HANDED_ALPHA
        )
        assert secondary_structure.start_sequence_code == 225
        assert secondary_structure.end_sequence_code == 233

    for sequence in structure.sequences.values():
        assert id(sequence.structure) == id(structure)
        assert sequence.id == 1
        assert sequence.source == SequenceSource.SEQRES



from pathlib import Path

import typer

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.tools.frames.list import list

app = typer.Typer()
app.command()(list)


# noinspection PyUnusedLocal
def test_frame_basic():

    path = path_in_test_data(__file__, "frames.nef")
    result = run_and_report(app, ["--in", path])

    EXPECTED = """\
        nef_nmr_meta_data  nef_molecular_system
    """

    assert_lines_match(EXPECTED, result.stdout)


# noinspection PyUnusedLocal
def test_frame_file_first():

    path = path_in_test_data(__file__, "frames.nef")

    result = run_and_report(app, [path])

    EXPECTED = """\
        nef_nmr_meta_data  nef_molecular_system
    """

    assert_lines_match(EXPECTED, result.stdout)


# noinspection PyUnusedLocal
def test_frame_file_first_and_select():

    path = path_in_test_data(__file__, "frames.nef")
    result = run_and_report(app, [path, "meta_data"])

    EXPECTED = """\
        nef_nmr_meta_data
    """

    assert_lines_match(EXPECTED, result.stdout)


def test_frame_no_file():

    path = Path("NO_SUCH_FILE")
    run_and_report(app, [path], expected_exit_code=1)


def test_frame_basic_verbose():

    path = path_in_test_data(__file__, "frames.nef")
    result = run_and_report(app, ["--in", path, "--verbose"])

    EXPECTED_VERBOSE = """
        entry test
            lines: 48 frames: 2 checksum: d6235a487cbdb33cb1ed5d2f5f3f2635 [md5]
        1. nef_nmr_meta_data
            category: nef_nmr_meta_data
            loops: 1 [lengths: 19]
            loop names: nef_program_script
            is nef frame: True
        2. nef_molecular_system
            category: nef_molecular_system
            loops: 1 [lengths: 13]
            loop names: nef_sequence
            is nef frame: True

    """

    assert_lines_match(EXPECTED_VERBOSE, result.stdout)



from enum import auto
from pathlib import Path
from textwrap import dedent, indent
from typing import List

import typer
from fyeah import f
from pynmrstar import Entry, Loop, Saveframe
from runstats import Statistics
from strenum import LowercaseStrEnum
from tabulate import tabulate

from nef_pipelines.lib.nef_lib import (
    UNUSED,
    SelectionType,
    create_nef_save_frame,
    get_frame_ids,
    read_entry_from_file_or_stdin_or_exit_error,
    select_frames,
)
from nef_pipelines.lib.peak_lib import frame_to_peaks
from nef_pipelines.lib.util import STDIN, exit_error
from nef_pipelines.tools.shifts import shifts_app

UPDATE_POLICY_HELP = (
    "how to update spectrum frame shift list to reflect the the calculated shift list"
)


class UpdatePolicy(LowercaseStrEnum):
    UPDATE = auto()
    UPDATE_UNDEFINED = auto()
    LEAVE = auto()


@shifts_app.command()
def average(
    in_path: Path = typer.Option(
        STDIN,
        "-i",
        "--in",
        metavar="|PIPE|",
        help="input to read NEF data from [- is stdin]",
    ),
    # selectors: str = typer.Argument(None, help="selectors to match shifts from [default is Spectra]"),
    # target: str = typer.Argument(None, help="shift list to put shifts in"),
    force: bool = typer.Option(False, help="if target peak list exists replace it"),
    frame_name: str = typer.Option("{entry_id}", help="the shift list to write to"),
    update_policies: List[UpdatePolicy] = typer.Option(
        [], "--update-policies", help=UPDATE_POLICY_HELP
    ),
    selectors: List[str] = typer.Argument(
        None, help="selectors for frames to average shifts from [default is all]"
    ),
):
    """- average chemical shifts from spectra [peaks] into a nef chemical shift frame [alpha]"""

    entry = read_entry_from_file_or_stdin_or_exit_error(in_path)

    selection_type = SelectionType.ANY
    if not selectors:
        selectors = "nef_nmr_spectrum"
        selection_type = SelectionType.CATEGORY

    frames = select_frames(entry, selectors, selection_type)

    if not update_policies:
        update_policies = [
            UpdatePolicy.UPDATE_UNDEFINED,
        ]

    _exit_if_bad_frame_types_selected(frames)

    entry = pipe(entry, frames, frame_name, force, update_policies=update_policies)

    print(entry)


def pipe(
    entry: Entry,
    frames: List[Saveframe],
    frame_name: str,
    force: bool,
    update_policies=(UpdatePolicy.UPDATE_UNDEFINED,),
) -> Entry:

    entry_id = entry.entry_id  # noqa: F841
    frame_name = f(frame_name)

    peaks = []
    for frame in frames:
        peaks.extend(frame_to_peaks(frame))
        _update_shift_list_name(frame, frame_name, update_policies)

    shifts = {}
    for peak in peaks:
        for shift in peak.shifts:
            shifts.setdefault(shift.atom, Statistics()).push(shift.value)

    results_frame = create_nef_save_frame("nef_chemical_shift_list", frame_name)

    loop = Loop.from_scratch("nef_chemical_shift")
    loop.add_tag(
        "chain_code sequence_code residue_name atom_name value value_uncertainty".split()
    )
    results_frame.add_loop(loop)

    for atom, stats in sorted(shifts.items()):
        stddev = stats.stddev() if len(stats) > 1 else UNUSED

        # filter on empty chain codes or sequence codes which are "" in the structures.Peak
        if not atom.residue.chain_code or not atom.residue.sequence_code:
            continue

        data = [
            {
                "chain_code": atom.residue.chain_code,
                "sequence_code": atom.residue.sequence_code,
                "residue_name": atom.residue.residue_name,
                "atom_name": atom.atom_name,
                "value": stats.mean(),
                "value_uncertainty": stddev,
            }
        ]
        loop.add_data(data)

    if frame_name in get_frame_ids(entry) and not force:
        exit_error(
            f"the frame {frame_name} already exists in the entry {entry.entry_id}, use --force to overwrite"
        )
    elif frame_name in get_frame_ids(entry) and force:
        old_saveframe = entry.get_saveframe_by_name(
            f"nef_chemical_shift_list_{frame_name}"
        )
        entry.remove_saveframe(old_saveframe)

    entry.add_saveframe(results_frame)

    return entry


def _update_shift_list_name(frame: Saveframe, shift_list_frame_name, update_policies):

    if (
        UpdatePolicy.UPDATE_UNDEFINED in update_policies
        and frame.get_tag("chemical_shift_list")[0] == UNUSED
    ):
        frame.add_tags(
            [
                (
                    "chemical_shift_list",
                    f"nef_chemical_shift_list_{shift_list_frame_name}",
                ),
            ],
            update=True,
        )
    else:
        msg = f"""
             the only supported update policy is currently {UpdatePolicy.UPDATE_UNDEFINED},
             you gave {', '.join(update_policies)} bug gary!
        """
        exit_error(msg)


def _exit_if_bad_frame_types_selected(frames):
    bad_frames = []
    for frame in frames:
        if frame.category not in [
            "nef_nmr_spectrum",
        ]:  # 'nef_chemical_shift_list']:
            bad_frames.append(frame)
    if bad_frames:
        bad_frame_table = []

        for frame in bad_frames:
            frame_name = frame.name
            frame_category = frame.category
            frame_name = frame_name[len(frame_category) :]
            if frame_name == "":
                frame_name = "**SINGLETON**"
            bad_frame_table.append([f"{frame_name}", f"[{frame.category}]"])

        msg = """
            Some of the selected frames are not chemical shift lists or spectra [peak lists]
            the bad frames and categories are:

        """
        msg = dedent(msg)
        msg += indent(
            tabulate(bad_frame_table, headers=["frame name", "category"]), "    "
        )
        msg += "\n"
        msg = f(msg)
        exit_error(msg)



from dataclasses import dataclass
from pathlib import Path
from typing import List

import typer
from pynmrstar import Entry, Loop, Saveframe
from uncertainties import ufloat

from nef_pipelines.lib.nef_frames_lib import NEF_PIPELINES_NAMESPACE
from nef_pipelines.lib.nef_lib import (
    NEF_RELAXATION_VERSION,
    UNUSED,
    create_nef_save_frame,
    get_frame_id,
    read_entry_from_file_or_stdin_or_exit_error,
)
from nef_pipelines.lib.shift_lib import IntensityMeasurementType
from nef_pipelines.lib.util import parse_comma_separated_options
from nef_pipelines.tools.fit import fit_app
from nef_pipelines.tools.fit.fit_lib import (
    _build_spectrum_peak_id_to_axis_atoms,
    _exit_if_isotope_axes_are_not_unique,
    _exit_if_isotope_frequencies_are_not_unique,
    _exit_if_no_frame_selectors,
    _exit_if_no_frequencies_found,
    _exit_if_no_series_frames_selected,
    _get_unique_isotope_frequencies,
    _select_relaxation_series_or_exit,
    _series_frame_to_data_id_to_spectrum_peak_id,
    _series_frame_to_id_series_data,
    _series_frame_to_spectrum_frames,
    frequencies_to_field_strength,
    spectra_to_isotope_axes,
    spectra_to_isotope_frequencies,
)


@dataclass
class RatioResult:
    value: float
    uncertainty: float


@fit_app.command()
def ratio(
    input: Path = typer.Option(
        None,
        "-i",
        "--in",
        metavar="NEF-FILE",
        help="read NEF data from a file instead of stdin",
    ),
    noise_level: float = typer.Option(
        ...,
        "-n",
        "--noise",
        help="noise level to use instead of value from spectra",
    ),
    data_type: IntensityMeasurementType = typer.Option(
        IntensityMeasurementType.HEIGHT, "-d", "--data-type", help="data type to fit"
    ),
    frames_selectors: List[str] = typer.Argument(None, help="select frames to fit"),
):
    """- calculate ratio of peak intensities with error propagation" [alpha]"""

    entry = read_entry_from_file_or_stdin_or_exit_error(input)

    frame_selectors = parse_comma_separated_options(frames_selectors)

    if not frames_selectors:
        series_frames = entry.get_saveframes_by_category("nef_series_list")
        if len(series_frames) == 1:
            frame_selectors = [
                series_frames[0].name,
            ]

    _exit_if_no_frame_selectors(frame_selectors)

    series_frames = _select_relaxation_series_or_exit(entry, frames_selectors)

    _exit_if_no_series_frames_selected(series_frames, frame_selectors)

    entry = pipe(entry, series_frames, noise_level, data_type)

    print(entry)


def pipe(
    entry: Entry,
    series_frames: List[Saveframe],
    noise_level,
    intensity_measurement_type: IntensityMeasurementType,
) -> Entry:

    for series_frame in series_frames:
        id_series_data = _series_frame_to_id_series_data(
            series_frame, NEF_PIPELINES_NAMESPACE, entry
        )

        id_xy_data = {
            id: (series_datum.variable_values, series_datum.values)
            for id, series_datum in id_series_data.items()
        }

        results = _ratio_calculation(id_xy_data, noise_level)

        results_frame = _ratio_results_as_frame(
            series_frame, NEF_PIPELINES_NAMESPACE, entry, results, noise_level
        )

        entry.add_saveframe(results_frame)

    return entry


def _ratio_calculation(id_xy_data, error):

    result = {}
    for id, xy_data in id_xy_data.items():
        xy_data = id_xy_data[id]

        x_data = xy_data[0]
        y_data = xy_data[1]

        on_index = [i for i, val in enumerate(x_data) if bool(val)][0]
        off_index = [i for i, val in enumerate(x_data) if not bool(val)][0]

        on_value = ufloat(y_data[on_index], error)
        off_value = ufloat(y_data[off_index], error)

        ratio = on_value / off_value

        result[id] = RatioResult(ratio.nominal_value, ratio.std_dev)

    return result


def _ratio_results_as_frame(
    series_frame,
    prefix,
    entry,
    results,
    noise_level,
):

    spectrum_frames = _series_frame_to_spectrum_frames(series_frame, prefix, entry)

    data_id_to_spectrum_peak_ids = _series_frame_to_data_id_to_spectrum_peak_id(
        series_frame, prefix
    )

    spectrum_peak_id_to_axis_atoms = _build_spectrum_peak_id_to_axis_atoms(
        spectrum_frames
    )

    isotope_axes = spectra_to_isotope_axes(spectrum_frames)

    _exit_if_isotope_axes_are_not_unique(isotope_axes, spectrum_frames, entry.entry_id)

    isotope_frequencies = spectra_to_isotope_frequencies(spectrum_frames)

    _exit_if_isotope_frequencies_are_not_unique(
        isotope_frequencies, spectrum_frames, entry.entry_id
    )

    isotope_frequencies = _get_unique_isotope_frequencies(isotope_frequencies)

    _exit_if_no_frequencies_found(isotope_frequencies, series_frame, entry)

    field_strength = frequencies_to_field_strength(isotope_frequencies)

    frame_id = get_frame_id(series_frame)

    result_frame = create_nef_save_frame(
        f"{prefix}_relaxation_list", f"{frame_id}_fitted"
    )

    result_frame.add_tag("version", NEF_RELAXATION_VERSION)
    result_frame.add_tag("experiment_type", series_frame.get_tag("experiment_type")[0])
    result_frame.add_tag("field_strength", field_strength)
    result_frame.add_tag("value_unit", UNUSED)
    result_frame.add_tag("relaxation_atom_id", UNUSED)
    result_frame.add_tag("ref_value", UNUSED)
    result_frame.add_tag("source", "experimental")
    comment = f"""
        noise_estimate {noise_level}
    """

    result_frame.add_tag("comment", comment)

    relaxation_loop = Loop.from_scratch(f"{prefix}_relaxation")
    result_frame.add_loop(relaxation_loop)

    RELAXATION_LOOP_TAGS = "index data_id data_combination_id value value_error".split()
    RELAXATION_LOOP_ATOM_TAGS = (
        "chain_code_{axis} sequence_code_{axis} residue_name_{axis} atom_name_{axis}"
    )

    all_tags = [
        *RELAXATION_LOOP_TAGS,
    ]
    for axis in range(1, len(isotope_axes) + 1):
        all_tags.append(RELAXATION_LOOP_ATOM_TAGS.format(axis=axis).split())

    relaxation_loop.add_tag(all_tags)

    data = []
    for index, (data_id, result) in enumerate(results.items(), start=1):
        data_row = {"index": index, "data_id": index, "data_combination_id": UNUSED}

        peak_atoms = []
        atom_key = next(iter(data_id_to_spectrum_peak_ids[data_id]))
        for axis_atoms in spectrum_peak_id_to_axis_atoms[atom_key].values():
            peak_atoms.extend(axis_atoms)

        for i, atom in enumerate(peak_atoms, start=1):
            residue = atom.residue
            data_row.update(
                {
                    f"chain_code_{i}": residue.chain_code,
                    f"sequence_code_{i}": residue.sequence_code,
                    f"residue_name_{i}": residue.residue_name,
                    f"atom_name_{i}": atom.atom_name,
                }
            )
        data.append(data_row)
        data_row.update(
            {
                "value": f"{result.value:.20f}",
                "value_error": f"{result.uncertainty:.20f}",
            }
        )

    relaxation_loop.add_data(data)

    return result_frame


# implementation thoughts
# 1. having a spectrometer frequency and field strength in the same construct is a pain, what do we dfeine as the
#    conversion ratio i took  alue in cavanagh and cross-checked it with from https://www.kherb.io/docs/nmr_table.html
#    which uses the codata value for 1H [but not for 15N and 13C which use the IAEA values]
# 2. if we have 1H and 15N and 13C spectrometer frequencies which should we use to
#    calculate the field strength? i chose the one with the highest gamma, but note my gamma ratios do not those in
#    https://www.kherb.io/docs/nmr_table.html exacly
# 3. do we need the value type should it be Nz for a 15N experiment, what are I and S defined as in this context
# 4. how do we determine the relaxation_atom_id shouldn't also be available on the series data or come from the series
#    data as it's the type of experiment and the experimenta series that determine this, currently choose lowest gamma!
# 5. the value doesn't have a type just a unit, should it?
# 6. the source maybe should be on the data series rather than the relaxation list -  it knoes what its is
#    the relaxation list doesn't need tow
# 7. the nef_relaxation doesn't leave room for other fit values e.g I or the baseline offset for a 3 point fit...
# 8. room for a comment on each fit would be good...
# 9. when would the index and data_id in a nef_relaxation_list be different?
# 10. we should say the value error is sigma
# 11. why are some of our constants capitalised and others not - we should be consistent
#     e.g. experimental | simulated | theoretical vs OnePhaseDecay | ExponentialDecay | InversionRecovery | other



from enum import auto
from pathlib import Path
from typing import List

import typer
from lazy_import import lazy_module
from strenum import LowercaseStrEnum

from nef_pipelines.lib.nef_lib import (
    read_entry_from_file_or_exit_error,
    read_or_create_entry_exit_error_on_bad_file,
)
from nef_pipelines.lib.sequence_lib import get_chain_code_iter
from nef_pipelines.lib.util import STDIN, exit_error, parse_comma_separated_options
from nef_pipelines.transcoders.nmrstar import import_app
from nef_pipelines.transcoders.nmrstar.importers.project_shortcuts import (
    BMRB_AUTO_MIRROR,
    BMRB_ITALY_URL_TEMPLATE,
    BMRB_JAPAN_URL_TEMPLATE,
    BMRB_URL_TEMPLATE,
    SHORTCUT_NAMES,
    SHORTCUTS,
)
from nef_pipelines.transcoders.nmrstar.nmrstar_lib import StereoAssignmentHandling

project_module = lazy_module("nef_pipelines.transcoders.nmrstar.importers.project")

app = typer.Typer()


class EntrySource(LowercaseStrEnum):
    FILE = auto()
    WEB = auto()
    AUTO = auto()


NO_CHAIN_START_HELP = """\
don't include the start chain link type on a chain for the first residue [linkage will be
middle] for the named chains. Either use a comma joined list of chains [e.g. A,B] or call this
option multiple times to set chain starts for multiple chains
"""

NO_CHAIN_END_HELP = """\
don't include the end chain link type on a chain for the last residue [linkage will be
middle] for the named chains. Either use a comma joined list of chains [e.g. A,B] or call this
option multiple times to set chain ends for multiple chains
"""

STEREO_HELP = """\
    how to handle stereo assignments the choices are:
    - ambiguous: assume all stereo assignments are ambiguous
    - as-assigned: use the stereo assignments as they are in the file
    - auto: use as assigned if some geminal stereo assignments are present, otherwise assume all are ambiguous
"""


FILE_PATH_HELP = f"""\
the file to read, if it is is one of the shortcuts {SHORTCUT_NAMES}, of the form bmr<NUMBER> or is just a number or
appears to be a url the program will attempt to fetch the entry from the bmrb or the web first before looking
for a file on disc unless this behaviour overridden by the --source option
"""


class Mirror(LowercaseStrEnum):
    AUTO = auto()
    USA = auto()
    JAPAN = auto()
    ITALY = auto()
    EUROPE = auto()
    ASIA = auto()


MIRROR_URLS = {
    Mirror.AUTO: BMRB_AUTO_MIRROR,
    Mirror.USA: BMRB_URL_TEMPLATE,
    Mirror.JAPAN: BMRB_JAPAN_URL_TEMPLATE,
    Mirror.ITALY: BMRB_ITALY_URL_TEMPLATE,
    Mirror.EUROPE: BMRB_ITALY_URL_TEMPLATE,
    Mirror.ASIA: BMRB_JAPAN_URL_TEMPLATE,
}

AUTO_MIRRORS = (
    MIRROR_URLS[Mirror.USA],
    MIRROR_URLS[Mirror.EUROPE],
    MIRROR_URLS[Mirror.ASIA],
)

URL_HELP = f"""
template for the bmrb url [default {BMRB_URL_TEMPLATE}],
note if a mirror is chosen than overrides a custom url
"""


@import_app.command()
def project(
    ctx: typer.Context,
    chain_codes: List[str] = typer.Option(
        [],
        "--chains",
        help="""chain codes to use for the imported chains, can be a comma separated list or can be called
                multiple times if no chain codes are provided they will be named A B C...""",
        metavar="<CHAIN-CODES>",
    ),
    no_chain_starts: List[str] = typer.Option(
        [], "--no-chain-start", help=NO_CHAIN_START_HELP
    ),
    no_chain_ends: List[str] = typer.Option(
        [], "--no-chain-end", help=NO_CHAIN_END_HELP
    ),
    source: EntrySource = typer.Option(
        EntrySource.AUTO, help="the source of the entry"
    ),
    use_author: bool = typer.Option(False, help="use author field for sequence codes"),
    stereo_mode: StereoAssignmentHandling = typer.Option("auto", help=STEREO_HELP),
    entry_name: str = typer.Option(
        None, help="a name for the entry (defaults to the bmr{entry_number})"
    ),
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--in",
        help="file to read NEF data from [- is stdin; defaults is stdin]",
    ),
    url_template: str = typer.Option(
        None,
        help=URL_HELP,
    ),
    list_shortcuts: bool = typer.Option(
        False, "--list-shortcuts", help="list the available shortcuts and exit"
    ),
    mirror: Mirror = typer.Option(None, help="select the mirror website to use"),
    verbose: bool = typer.Option(False, help="print verbose output"),
    timeout: int = typer.Option(10, help="timeout (seconds)  for http responses"),
    file_paths: List[str] = typer.Argument(None, help=FILE_PATH_HELP),
):
    """- convert as much as possible from an NMR-STAR file to NEF [shifts & sequences] [alpha]"""

    if list_shortcuts:
        project_module._list_shortcuts_and_exit()

    file_paths = parse_comma_separated_options(file_paths)
    file_paths = [
        SHORTCUTS[file_path.upper()] if file_path.upper() in SHORTCUTS else file_path
        for file_path in file_paths
    ]

    if not file_paths:
        typer.echo(ctx.get_help())
        exit_error("missing file path argument")

    if not url_template and not mirror:
        mirror = Mirror.AUTO

    if url_template and mirror:
        exit_error("cannot set both a url template and a mirror")
    elif mirror == Mirror.AUTO:
        url_templates = AUTO_MIRRORS
    elif mirror:
        if mirror not in MIRROR_URLS:
            msg = f"mirror must be one of {MIRROR_URLS.keys()}, i got {mirror}"
            exit_error(msg)
        url_templates = [
            MIRROR_URLS[mirror],
        ]
    else:
        url_templates = [
            MIRROR_URLS[Mirror.USA],
        ]

    for file_path in file_paths:
        chain_codes = parse_comma_separated_options(chain_codes)
        chain_codes = get_chain_code_iter(chain_codes)

        no_chain_starts = parse_comma_separated_options(no_chain_starts)
        if not no_chain_starts:
            no_chain_starts = [False]

        no_chain_ends = parse_comma_separated_options(no_chain_ends)
        if no_chain_ends:
            no_chain_ends = [False]

        is_bmrb = False
        if source in (EntrySource.AUTO, EntrySource.WEB):
            urls, is_bmrb = project_module._get_path_as_url_or_none(
                file_path, url_templates
            )

            exit_on_error = True if source == EntrySource.WEB else False

            possible_entry = project_module._get_bmrb_entry_from_web_or_none(
                urls, exit_on_error, verbose=verbose, timeout=timeout
            )

            nmrstar_entry = project_module._parse_text_to_star_or_none(possible_entry)

        if (
            source == EntrySource.AUTO and not nmrstar_entry
        ) or source == EntrySource.FILE:
            nmrstar_entry = read_entry_from_file_or_exit_error(file_path)

        if not nmrstar_entry:
            msg = f"could not read entry from {file_path}"
            exit_error(msg)

        entry_name = nmrstar_entry.entry_id if not entry_name else entry_name
        entry_name = f"bmr{entry_name}" if is_bmrb else entry_name
        nef_entry = read_or_create_entry_exit_error_on_bad_file(
            input, entry_name=entry_name
        )

        entry = project_module.pipe(
            nef_entry,
            nmrstar_entry,
            chain_codes,
            no_chain_starts,
            no_chain_ends,
            use_author,
            stereo_mode,
            file_path,
        )

        print(entry)
